// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: JavaClassFile.h,v 1.14 1998/07/21 21:09:03 nickm Exp $
#ifndef _JavaClassFile_h
#define _JavaClassFile_h
#pragma interface

#include "JavaConstant.h"
#include "JavaFieldInfo.h"
#include "JavaMethodInfo.h"
#include "JavaMethodSignature.h"
#include "JavaAccessFlags.h"
#include "PJavaClassInfo.h"
#include "PrimitiveMethods.h"
#include "unicode_string.h"
#include <time.h>
#include <vector>
#include <map>
#include <set>
class CJavaSourceFileAttribute;
class CCompiler;
class CJavaMethodInfo;
class ostream;
class istream;
class ifstream;

typedef map<CJavaFieldSignature, const CJavaFieldInfo*, 
    less<CJavaFieldSignature> > FieldTable;
typedef map<CJavaMethodSignature, const CJavaMethodInfo*, 
    less<CJavaMethodSignature> > MethodTable;
typedef set<unicode_string, less <unicode_string> > UnicodeStringSet;

//
//  Class name : CJavaClassFile
//  Description : This class tries to encapsulate the Java class file format
//    as specified in the Java virtual machine documentation.
//    As such, this isn't designed to be optimized for any specific usage
//    other than streaming to and from a java class file.
//
class CJavaClassFile {
public:
 virtual ~CJavaClassFile();
  
  static CJavaClassFile* ParseFromString(const string& fileBuffer,
					 bool interfaceOnly = false);
  static CJavaClassFile* ParseFromStream(istream& inStream,
					 bool interfaceOnly = false);
  static CJavaClassFile* ParseFromFile(ifstream& inStream,
				       bool interfaceOnly = false);
  static u16 ReadJavaU2(string::const_iterator& bufferPointer);
  static u32 ReadJavaU4(string::const_iterator& bufferPointer);
  
  CJavaAccessFlags GetAccessFlags() const { return fAccessFlags; }

  typedef vector<CJavaConstant*> ConstantPool;
  const CJavaConstant* LookupConstant(JavaConstantIndex index) const;
  unicode_string GetClassName() const { return fThisClassName; }
  unicode_string GetSuperclassName() const { return fSuperclassName; }
  typedef deque<unicode_string> InterfaceList;
  unsigned long CountInterfaces() const;
  InterfaceList::const_iterator GetInterfaceBegin() const;
  InterfaceList::const_iterator GetInterfaceEnd() const;
  bool SameParents(const CJavaClassFile& other) const;

  unicode_string GetShortName() const;
  unicode_string GetPackageName() const;
  unsigned long GetCompilerVersion() const;
  bool LookupFieldSignature(JavaConstantIndex index,
			    unicode_string& setClassName,
			    CJavaFieldSignature& setField) const;
  bool LookupMethodSignature(JavaConstantIndex index,
			     unicode_string& setClassName,
			     CJavaMethodSignature& setMethod) const;
  bool LookupInterfaceSignature(JavaConstantIndex index,
				unicode_string& setClassName,
				CJavaMethodSignature& setMethod) const;
  bool LookupClassName(JavaConstantIndex index, unicode_string& setName) const;

  const CJavaFieldInfo* LookupField(const unicode_string& name) const;
  const CJavaFieldInfo* LookupField(const CJavaFieldSignature& name) const;
  const CJavaMethodInfo*
                    LookupMethod(const CJavaMethodSignature& method) const;
  const CJavaMethodInfo*
           LookupMethodIgnoreReturn(const CJavaMethodSignature& method) const;
  bool IsDeprecated() const;

  void DisassembleClass(ostream& toStream) const;
  void Compile(ostream& toStream);
  JavaConstantIndex AddLongConstant(JavaULong doubleLong);
  JavaConstantIndex AddIntegerConstant(JavaUInt integerValue);
  JavaConstantIndex AddFloatConstant(float value);
  JavaConstantIndex AddDoubleConstant(double value);
  JavaConstantIndex AddAscizConstant(const unicode_string& value);
  JavaConstantIndex AddAscizConstant(const string& value);
  JavaConstantIndex AddStringConstant(JavaConstantIndex ascizIndex);
  JavaConstantIndex AddClassConstant(const unicode_string& name);
  JavaConstantIndex AddFieldConstant(const unicode_string& className,
	     const unicode_string& fieldName, const CJavaTypeSignature& type);
  JavaConstantIndex AddMethodConstant(const unicode_string& className,
				      const CJavaMethodSignature& method);
  JavaConstantIndex AddInterfaceConstant(const unicode_string& className,
					 const CJavaMethodSignature& method);
  JavaConstantIndex AddNameTypeConstant(const unicode_string& name,
					const unicode_string& typeString);

  CJavaSourceFileAttribute* GetSourceFile() const;

  bool IsParameterized() const 
    { return (fPClassInfo != 0) && fPClassInfo->IsParameterized(); }
  bool HasPInfo() const 
    { return (fPClassInfo != 0); }
  CPJavaClassInfo& GetPInfo() const 
    { assert(fPClassInfo); return *fPClassInfo; }
  bool ParsePJavaClassInfo(const string & str) 
    { fPClassInfo = new CPJavaClassInfo; 
      return fPClassInfo->ParseFromString(*this,str); }
  const CJavaTypeSignature &GetType() const
    { if (!fType) ((CJavaClassFile*)this)->SetType(); return *fType; }

  bool HasParameterNamed(const unicode_string& name) const;

  void SetDate(time_t t) { fDate = t; }

protected:
  friend class CCompiler;
  friend class CPJavaClassInfo;
  friend CJavaClassFile* CreateClassForType(const CJavaTypeSignature&, 
					    const unicode_string&);
  CJavaClassFile();
  void InsertMethod(CJavaMethodInfo* methodInfo);
  JavaConstantIndex InsertConstant(CJavaConstant* adoptConstant);
  bool SeemsParameterized() const;
private:
  void SetType();

  unsigned long fVersion;
  ConstantPool fConstants;
  CJavaAccessFlags fAccessFlags;
  unicode_string fThisClassName;
  unicode_string fSuperclassName;
  InterfaceList fInterfaces;
  FieldTable fFields;
  MethodTable fMethods;
  bool fDeprecated;

  CJavaSourceFileAttribute* fSourceFile;
  CPJavaClassInfo* fPClassInfo;
  CJavaTypeSignature *fType;
  
  time_t fDate;
};

#endif


