#pragma implementation

#include "mapped_file.h"
#include <iostream.h>
#include "find_leaks.h"

#ifdef HAVE_SYSCTL
#if !defined(CTL_HW) || !defined(HW_PAGESIZE)
#undef HAVE_SYSCTL
#endif
#endif

extern "C" {
  #include <errno.h>
  #include <sys/stat.h>
  #include <unistd.h>
  #include <string.h>
  #include <stdio.h>
  #include <assert.h>
#ifdef HAVE_MMAP
  #include <sys/mman.h>
  #ifndef MAP_VARIABLE
     #define MAP_VARIABLE 0
  #endif
  #ifdef HAVE_SYSCTL
    #include <sys/param.h>
    #include <sys/sysctl.h>
  #else
    const size_t kPageSize = ::sysconf(_SC_PAGESIZE);
  #endif
#endif
}

// 
//  Method name: (constructor) 
//  Description: Pulls a portion of a file beginning at pos, of size len,
//      into memory.
//
mapped_file::mapped_file(int fd, off_t pos, size_t len)
{
  REF_PTR(this);
  unexpected_eof = false;
  map_file(fd, pos, len);
}

// 
//  Method name: (constructor)
//  Description: Pulls an entire file into memory.
//
mapped_file::mapped_file(int fd)
{
  REF_PTR(this);
  struct stat st;
  assert(fd >= 0);
  fstat(fd,&st);
  unexpected_eof = false;
  map_file(fd,0,st.st_size);
}


// 
//  Method name: (destructor)
//
mapped_file::~mapped_file()
{
  UNREF_PTR(this);
}

// 
//  Method name: map_file
//  Description: Protected method reads a given file into memory, by hook
//      or crook. (Where hook=mmap(2) and crook=read(2).)
//
void 
mapped_file::map_file(int fd, off_t pos, size_t len)
{
  assert(fd >= 0);
  /* Always true. assert(pos >= 0 && len >= 0); */
  struct stat st;
  assert(fd >= 0);
  int res = fstat(fd,&st);
  assert(res == 0);
  valid = false;  
#ifdef HAVE_MMAP
  #ifdef HAVE_SYSCTL
    static size_t kPageSize = 0;
    if (kPageSize == 0) {
      unsigned int len = 0;
      int mib[2]={0,0};
      int result;
      len = sizeof(result);
      mib[0] = CTL_HW;
      mib[1] = HW_PAGESIZE;
      sysctl(mib, 2, &result, &len, NULL, 0);
      kPageSize = result;
    }
  #endif
  size_t map_len = len + (pos%kPageSize);
  off_t round_pos = (pos/kPageSize)*kPageSize;
  buffer = mmap(NULL, map_len, PROT_READ|PROT_WRITE, //v2
  		MAP_PRIVATE|MAP_VARIABLE, fd, round_pos);
  if ((long) buffer == -1) {
    perror("guavac"); 
    valid = false;
  } else {
    valid = true;
    str.assign(((char*)buffer)+(pos%kPageSize),len);
    munmap((void*) buffer, map_len);
  }
#endif
  if (!valid) {
    char buffer[1024]; 
    if (::lseek(fd,pos,SEEK_SET) >= 0) {
      ssize_t result = 0;
      while(len > 0 && result != -1) {
	size_t l = (1024 > len) ? len : 1024;
	result = ::read(fd,buffer,l);
	if (result > 0) {
	  str.append(buffer,result);
	  len -= result;
	} else if (result == 0 && len > 0) {
	  result = -1;
	  unexpected_eof = true;
	}
      }
      valid = (len == 0 && result != -1);
    } else {
      valid = false;
    }
  }
  my_errno = errno;
}

void
mapped_file::print_error() 
{
  if (!valid) {
    if (unexpected_eof) {
      cerr << "Error: Unexpected EOF on read." << endl;
    } else {
      cerr << "Error: " << strerror(my_errno) << endl;
    }
  }
}

