// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: Expression.C,v 1.54 1998/10/09 19:19:57 nickm Exp $
#pragma implementation
#include "find_leaks.h"

#include "JavaTypeSignature.h"
#include <map>
#include <functional>
#include "Expression.h"
#include "parser_decls.h"
#include "parser.h"  // for token values
#include "JavaFieldSignature.h"
#include "JavaClassFile.h"
#include "Compiler.h"
#include "CompileError.h"
#include "JavaMethodInfo.h"
#include "JavaCodeAttribute.h"
#include "CompileContext.h"
#include "Srcstream.h"
#include "Instantiator.h"
#include "JavaConstant.h"
#include "PJavaClassInfo.h"
#include <iomanip.h>
#include <algorithm>
#include <stdio.h>
extern "C" {
  #include <math.h>
}
#ifndef OLDGCC
using namespace std;
#endif

#if defined(HAVE_IEEEFP_H)
#include <ieeefp.h>
#endif

//=============================== CExpression ================================

//
//  Method name : CExpression
//  Description : Protected constructor.  Used by children to construct an
//    expression instance that knows its type and knows whether or not it is
//    an lvalue-type expression that can be assigned to.
//    The 'Void' type is used as a placeholder to say 'I don't know what type
//    this is yet.'  Because of use-before-declaration in Java, types can't
//    be determined in the first compilation pass, so I stick this type in
//    as a placeholder until the type-check pass.
//
//    You could argue that it's poor encapsulation to grab the line number
//    from the global yylineno variable like I'm doing here.  I'd probably
//    agree with you, but the alternative is to pass the line number into
//    the constructors of every subclass of CExpression and then pass it into
//    this constructor, which would really make for tedious code.
//    Maybe it will be replaced some day.
//    "Aqua sea foam shame." -- Nirvana, 'All Apologies'
//
CExpression::CExpression(const CJavaTypeSignature& type, bool isLValue)
  : fType(type), fTypeHasBeenEvaluated(false), fLValue(isLValue), 
    fLineNumber(yylineno), fPrecedence(-100), fLValAlias(false)
{
}

//
//  Method name : CExpression
//  Description : Copy constructor.
//
CExpression::CExpression(const CExpression& source)
  : fType(source.fType),
    fTypeHasBeenEvaluated(source.fTypeHasBeenEvaluated),
    fLValue(source.fLValue),
    fLineNumber(source.fLineNumber),
    fPrecedence(source.fPrecedence),
    fLValAlias(source.fLValAlias)   
{
}

//
//  Method name : ~CExpression
//  Description : Destructor
//
CExpression::~CExpression()
{
}

//
//  Method name : EvaluateType
//  Description : This is used during the second compilation pass to tell an
//    expression to determine its type (if it hasn't done so already) and
//    pass it back to the caller through the 'setType' parameter.  This
//    call is also used to simplify constant-valued expressions (e.g. 2 + 3),
//    so the a pointer to the relevant expression is passed to this call.  If
//    a constant-folded substitution needs to be made, the original referent
//    of the pointer will be deleted and a replacement will be provided.
//    The other parameters are just to give the context so it can search for
//    fields and classes, etc.  If successful, it returns 0, otherwise it
//    returns a new error object to the user.
//
CCompileError*
CExpression::EvaluateType(CExpression*& expressionPointer,
	  CCompileContext& context, CJavaTypeSignature& setType, bool asLValue)
{
  assert(expressionPointer != 0);
  CCompileError* error = 0;

  // Resolve type.
  CExpression* setExp = 0;
  error = expressionPointer->Resolve(context,setExp);
  if (error)
    return error;
  if (setExp) {
    delete expressionPointer;
    expressionPointer = setExp;
  }
 
  if (!asLValue) {
    expressionPointer->fLValue = false;
  }
  if (!context.IsReachable()) {
    error = expressionPointer->MakeError("Expression cannot be reached");
  } else if (expressionPointer->fType == CJavaTypeSignature::kVoid && 
	     !expressionPointer->fTypeHasBeenEvaluated) {
    error = expressionPointer->HandleEvaluateType(context,
						  expressionPointer->fType);
    expressionPointer->fTypeHasBeenEvaluated = true;
  }
  unicode_string className;
  if (error == 0 && expressionPointer->fType != CJavaTypeSignature::kNullType
      && expressionPointer->fType.GetBaseClassName(className)) {
    // Check and fix all types
    error = context.GetCompiler().CheckType(expressionPointer->fType,
					    context.GetClass(),
					    context.GetSource());
    if (error) error->SetLine(expressionPointer->fLineNumber); 
    expressionPointer->fType = 
      context.GetCompiler().FixType(expressionPointer->fType,
				    context.GetSource());
    if (error == 0) {
      if (expressionPointer->fType.IsParameter()) {
        assert(context.GetClass().HasParameterNamed(className));
      } else {
	const CJavaClassFile* classFile = 
	  context.GetCompiler().ClassOfType(expressionPointer->fType,
					    context.GetSource());
	if (classFile == 0) {
 	  unicode_string errorString = 
	    ::StringToUnicode("Unable to find class ");
	  errorString += className;
	  error = expressionPointer->MakeError(errorString);
	} else if (!(classFile->GetClassName() == className)) {
	  expressionPointer->fType.SetBaseClassName(classFile->GetClassName());
	}
      }
    }
  }
  if (error == 0) {
    if (!asLValue && !expressionPointer->IsLiteral()) {
      CExpression* temporaryExpression = 0;
      error = expressionPointer->CreateConstantExpression(temporaryExpression,
							  context);
      if (temporaryExpression != 0 && error == 0) {
	temporaryExpression->InheritLineNumber(*expressionPointer);
	delete expressionPointer;
	expressionPointer = temporaryExpression;
      }
    }
    setType = expressionPointer->fType;
  }
  return error;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.   It behaves
//    like EvaluateType in terms of parameters and return values.
//
CCompileError*
CExpression::HandleEvaluateType(CCompileContext& context,
				CJavaTypeSignature& setType)
{
  // This implementation is provided for children that should always have
  // their type known at the time of the first pass (like literals) ... since
  // EvaluateType knows their type, this should never be called.
  assert(0);
  return 0;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CExpression::CreateConstantExpression(CExpression*& result,
				      CCompileContext& context) const
{
  result = 0;
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CExpression::Unparse(srcstream &sOut,
		     CCompileContext& context,
		     bool isStatement)
{
  // This method should always be overridden.
  assert(0);
  return 0;
}

//
//  Method name : UnparseStore
//  Description : Writes the Java code for setting "this" to rValue to sOut.
//
CCompileError*
CExpression::UnparseStore(srcstream &sOut,
			  CCompileContext& context,
			  CExpression& rValue,
			  bool isStatement)
{
  CCompileError* error = Unparse(sOut,context);
  if (error == 0) {
    sOut.out() << " = ";
    error = rValue.Unparse(sOut,context);
  }
  return error;
}

//
//  Method name : MakeError
//  Description : This is a protected convenience function so the subclasses
//    of CExpression can create a compile error from a string.
//
CCompileError*
CExpression::MakeError(const string& message) const
{
  return new CCompileError(::StringToUnicode(message), fLineNumber);
}

//
//  Method name : MakeError
//  Description : This is a protected convenience function so the subclasses
//    of CExpression can create a compile error from a string.
//
CCompileError*
CExpression::MakeError(const unicode_string& message) const
{
  return new CCompileError(message, fLineNumber);
}

//
//  Method name : UnaryPromote
//  Description : This static member function can be used by children to try
//    to promote a pointer to a short or byte expression to int by applying a
//    cast operation to it.  See language spec section 3.5
//    Returns true if the expression has an integer type after the call.
//
bool
CExpression::UnaryPromote(CExpression*& promotePointer,
			  CCompileContext& context)
{
  assert(promotePointer != 0);
  bool result;
  switch (promotePointer->GetType().GetBaseType()) {
  case CJavaTypeSignature::Short:
  case CJavaTypeSignature::Byte:
  case CJavaTypeSignature::Character:   // for compatibility .. violates S3.5
    {
      CCompileError* error = CCastExpression::CastPointer(promotePointer,
				      CJavaTypeSignature::kInteger, context);
      assert(error == 0);
    }
  case CJavaTypeSignature::Integer:
    result = true;
    break;
  default:
    result = false;
  }
  return result;
}

//
//  Method name : BinaryPromote
//  Description : Takes two expression pointers and tries to promote them
//    to a matching numeric type through cast wrapper objects.  If the
//    two expressions are of matching types by the end of this call, true
//    is returned (even if it didn't need to do anything), otherwise false
//    is returned.  Read section 3.6 of the language spec for more information.
//
bool
CExpression::BinaryPromote(CExpression*& left, CExpression*& right,
			   CCompileContext& context)
{
  assert(left != 0 && right != 0);
  CCompileError* error = 0;
  CJavaTypeSignature leftType = left->GetType();
  CJavaTypeSignature rightType = right->GetType();
  bool result = false;
  if (leftType == CJavaTypeSignature::kDouble) {
    result = rightType == CJavaTypeSignature::kDouble;
    if (!result) {
      if (rightType.IsNumeric()) {
	error = CCastExpression::CastPointer(right, leftType, context);
	assert(error == 0);
	result = true;
      }
    }
  } else if (rightType == CJavaTypeSignature::kDouble) {
    if (leftType.IsNumeric()) {
      error = CCastExpression::CastPointer(left, rightType, context);
      assert(error == 0);
      result = true;
    }
  } else if (leftType == CJavaTypeSignature::kFloat) {
    result = rightType == CJavaTypeSignature::kFloat;
    if (!result) {
      if (rightType.IsNumeric()) {
	error = CCastExpression::CastPointer(right, leftType, context);
	assert(error == 0);
	result = true;
      }
    }
  } else if (rightType == CJavaTypeSignature::kFloat) {
    if (leftType.IsNumeric()) {
      error = CCastExpression::CastPointer(left, rightType, context);
      assert(error == 0);
      result = true;
    }
  } else if (leftType == CJavaTypeSignature::kLong) {
    result = rightType == CJavaTypeSignature::kLong;
    if (!result) {
      if (rightType.IsNumeric()) {
	error = CCastExpression::CastPointer(right, leftType, context);
	assert(error == 0);
	result = true;
      }
    }
  } else if (rightType == CJavaTypeSignature::kLong) {
    if (leftType.IsNumeric()) {
      error = CCastExpression::CastPointer(left, rightType, context);
      assert(error == 0);
      result = true;
    }
  } else {
    result = leftType.IsNumeric() && rightType.IsNumeric();
    if (result) {
      if (!(leftType == CJavaTypeSignature::kInteger)) {
	error = CCastExpression::CastPointer(left,
				 CJavaTypeSignature::kInteger, context);
      }
      if (error == 0 && !(rightType == CJavaTypeSignature::kInteger)) {
	error = CCastExpression::CastPointer(right,
			     CJavaTypeSignature::kInteger, context);
      }
    }
  }
  return result;
}

//
//  Method name : SetType
//  Description : Protected method to allow an expression to specify its
//    compile-time type.
//
void
CExpression::SetType(const CJavaTypeSignature& type)
{
  fType = type;
}

//
//  Method name : InheritLineNumber
//  Description : This method tells an expression that it should use the
//    source line number of another expression as its own.
//
void
CExpression::InheritLineNumber(const CExpression& from)
{
  fLineNumber = from.fLineNumber;
}

//
//  Method name : IsLiteral
//  Description : This method tells the caller whether this expression has a
//    constant value that could be used in a final interface method, a 'case'
//    label, etc.  This is the default implementation, which indicates that
//    the value is not constant.
//
bool
CExpression::IsLiteral() const
{
  return false;
}

//
//  Method name : MayBeType
//  Description : Thanks to the gymnastics we perform in order to use single
//    brackets for our IType symbols, we need to have a method in order to
//    tell us whether -- from a syntactical standpoint alone -- this expression
//    could possible be reinterpreted as a plausible type.
//
//  "Oh what a tangled web we weave
//    When context and syntax interleave."
//
bool
CExpression::MayBeType() const
{
  return false;
}

//
//  Method name : ToType
//  Description : Returns a new CJavaTypeSignature for the type encoded
//     by this AmbiguousAccess.  Returns NULL if this is impossible.
//
CJavaTypeSignature* 
CExpression::ToType() const
{
  return 0;
}

//
//  Method name : UnparseWrap
//  Description : This static method unparses "wrapped" to sOut, in context,
//    putting it in the wrapper appropriate to wrapAs.
//
CCompileError* 
CExpression::UnparseWrap(const CJavaTypeSignature &wrapAs,
			 CExpression& wrapped,
			 srcstream &sOut,
			 CCompileContext &context)
{
  CCompileError *err = 0;
  if (wrapAs.GetBaseType() == CJavaTypeSignature::Boolean) {
    sOut.out() << ::UnicodeToUTF(wrapAs.GetWrapperName())
	       << ".Wrap(";
    err = wrapped.Unparse(sOut,context);
    sOut.out() << ')';
  } else {
    sOut.out() << "(Object) new " << ::UnicodeToUTF(wrapAs.GetWrapperName())
	       << "(";
    err = wrapped.Unparse(sOut,context);
    sOut.out() << ")";
  }
  return err;
}

//
//  Method name : Resolve
//  Description : Attempts to resolve any ambiguities in this expression.
//      Does not attempt to resolve ambiguities in subexpressions.  On
//      failure, returns an error message.  If the expresion-class changes,
//      a new expression is written into setExpression.
//
CCompileError* 
CExpression::Resolve(CCompileContext& context,
		     CExpression*& setExpression)
{
  return 0;
}

//
//  Method name : UnparseArgument
//  Description : This static methods unparses "wrapped" to sOut, in context,
//    as an argument to method which expects an argument of "argType" before
//    translation, and "transType" after.
//
CCompileError* 
CExpression::UnparseArgument(const CJavaTypeSignature &argType,
			     const CJavaTypeSignature &transType,
			     CExpression& wrapped,
			     srcstream &sOut,
			     CCompileContext &context)
{
  CCompileError *err = 0;

  bool needCast = ! context.GetCompiler().SameType(wrapped.GetType(),
						   transType,
						   context.GetSource());
  bool needWrap = (needCast && 
		   wrapped.GetType().IsPrimitive() &&
		   ! transType.IsPrimitive());

  if (needWrap) {
    err = CExpression::UnparseWrap(argType, wrapped, sOut, context);
  } else if (needCast) {
    sOut.out() << "(";
    transType.Unparse(sOut);
    sOut.out() << ") ";
    err = wrapped.Unparse(sOut,context);
  } else {
    err = wrapped.Unparse(sOut,context);
  }
  return err;
}

//
//  Method name : UnparseWClassCode
//  Description : This static methods writes to sOut the code needed to
//    obtain, within 'context', a WObject for type.  It adds type to
//    the instantiation list if it's fully instantiated.
//
void
CExpression::UnparseWClassCode(const CJavaTypeSignature &type,
			       CCompileContext &context,
			       srcstream& sOut)
{
  assert(type.IsParameterized());

  CJavaTypeSignature t = context.GetCompiler().FixType(type,
						       context.GetSource());
  const CJavaClassFile &Class = context.GetClass();
  bool sameType = context.GetCompiler().SameType(t,Class.GetType(),
						 context.GetSource());

  if (sameType) {
    sOut.out() << "$W";
  } else if (type.IsInstantiable()) {
    context.GetCompiler().GetIList().AddType(t, context.GetSource());
    t.UnparseInstantiation(sOut,context.GetCompiler());
    sOut.out() << ".getInstance()";
  } else {
    unicode_string getWhere = Class.GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::GetWhere, type);
    sOut.out() << "$W." << ::UnicodeToUTF(getWhere) << "()";
  }
}

//============================= COrdinalLiteral ===============================

//
//  Method name : COrdinalLiteral
//  Description : Constructs an integer literal value with the provided type
//    signature.
//
COrdinalLiteral::COrdinalLiteral(JavaUInt integerLiteral,
				 const CJavaTypeSignature& type)
  : CExpression(type)
{
  fOrdinalLiteral = integerLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs an integer literal value.
//
COrdinalLiteral::COrdinalLiteral(JavaUInt integerLiteral)
  : CExpression(CJavaTypeSignature::Integer)
{
  fOrdinalLiteral = integerLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs an integer literal value.
//
COrdinalLiteral::COrdinalLiteral(JavaSInt integerLiteral)
  : CExpression(CJavaTypeSignature::Integer)
{
  fOrdinalLiteral = integerLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a two-byte literal value.  If the 'isChar' flag
//    is 'true', then this is a character literal, otherwise it is a short
//    constant value.
//
COrdinalLiteral::COrdinalLiteral(JavaUChar characterLiteral, bool isChar)
  : CExpression(isChar ? CJavaTypeSignature::Character :
		CJavaTypeSignature::Short)
{
  fOrdinalLiteral = characterLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a signed, two-byte integer literal.
//
COrdinalLiteral::COrdinalLiteral(JavaSShort shortLiteral)
  : CExpression(CJavaTypeSignature::Short)
{
  fOrdinalLiteral = shortLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a constant byte value.
//
COrdinalLiteral::COrdinalLiteral(JavaUByte integerLiteral)
  : CExpression(CJavaTypeSignature::Byte)
{
  fOrdinalLiteral = integerLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a constant byte value.
//
COrdinalLiteral::COrdinalLiteral(JavaSByte integerLiteral)
  : CExpression(CJavaTypeSignature::Byte)
{
  fOrdinalLiteral = integerLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a boolean literal value.
//
COrdinalLiteral::COrdinalLiteral(bool boolLiteral)
  : CExpression(CJavaTypeSignature::Boolean)
{
  fOrdinalLiteral = (boolLiteral ? 1 : 0);
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a double-long integer literal value.
//
COrdinalLiteral::COrdinalLiteral(JavaULong longLiteral)
  : CExpression(CJavaTypeSignature::LongInteger)
{
  fOrdinalLiteral = longLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Constructs a double-long integer literal value.
//
COrdinalLiteral::COrdinalLiteral(JavaSLong longLiteral)
  : CExpression(CJavaTypeSignature::LongInteger)
{
  fOrdinalLiteral = longLiteral;
}

//
//  Method name : COrdinalLiteral
//  Description : Copy constructor.
//
COrdinalLiteral::COrdinalLiteral(const COrdinalLiteral& source)
  : CExpression(source), fOrdinalLiteral(source.fOrdinalLiteral)
{
}

//
//  Method name : COrdinalLiteral
//  Description : Destructor
//
COrdinalLiteral::~COrdinalLiteral()
{
}


//
//  Method name : AddToConstantPool
//  Description : This method is used to add this value into the constant
//    pool of the provided class.  It returns the pool index that was used.
//
JavaConstantIndex
COrdinalLiteral::AddToConstantPool(CJavaClassFile& inClass) const
{
  JavaConstantIndex index;
  if (GetType() == CJavaTypeSignature::kLong) {
    index = inClass.AddLongConstant(fOrdinalLiteral);
  } else {
    index = inClass.AddIntegerConstant(fOrdinalLiteral);
  }
  return index;
}

//
//  Method name : IsLiteral
//  Description : This method tells the caller whether this expression has a
//    constant value that could be used in a final interface method, a 'case'
//    label, etc.
//
bool
COrdinalLiteral::IsLiteral() const
{
  return true;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
COrdinalLiteral::CreateConstantExpression(CExpression*& result,
					  CCompileContext& context) const
{
  result = Clone();
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
COrdinalLiteral::Unparse(srcstream &sOut,
			 CCompileContext& context,
			 bool isStatement)
{
  sOut.ToLine(fLineNumber,false);
  switch(fType.GetBaseType()) 
    {
    case CJavaTypeSignature::Byte:
      sOut.out() << "(byte)" << ((int) (JavaSByte) fOrdinalLiteral);
      break;      
    case CJavaTypeSignature::Short:
      sOut.out() << "(short)" << ((JavaSShort) fOrdinalLiteral);
      break;
    case CJavaTypeSignature::Integer:
      sOut.out() << ((JavaSInt) fOrdinalLiteral);
      break;      
    case CJavaTypeSignature::LongInteger:
      sOut.out() << ((JavaSLong) fOrdinalLiteral) << 'L';
      break;
    case CJavaTypeSignature::Character:
      sOut.out() << '\'' << ::CharLitToString(fOrdinalLiteral) << '\'';
      break;     
    case CJavaTypeSignature::Boolean:
      sOut.out() << (fOrdinalLiteral ? "true" : "false");
      break;
    default:
      assert(0);
      break;
    }
    
  return 0;
}


//============================= CFloatLiteral ===============================

//
//  Method name : CFloatLiteral
//  Description : Creates a 32 bit floating-point literal.
//
CFloatLiteral::CFloatLiteral(float floatLiteral, string* optAdoptString)
  : CExpression(CJavaTypeSignature::Float), fValue(floatLiteral),
    fString(optAdoptString)
{
}

//
//  Method name : CFloatLiteral
//  Description : Creates a 64 bit floating-point literal.
//
CFloatLiteral::CFloatLiteral(double doubleLiteral, string* optAdoptString)
  : CExpression(CJavaTypeSignature::Double), fValue(doubleLiteral),
    fString(optAdoptString)
{
}

//
//  Method name : CFloatLiteral
//  Description : Copy constructor.
//
CFloatLiteral::CFloatLiteral(const CFloatLiteral& source)
  : CExpression(source), fValue(source.fValue)
{
  if (source.fString)
    fString = new string(*source.fString);
  else
    fString = 0;
}

//
//  Method name : ~CFloatLiteral
//  Description : Destructor
//
CFloatLiteral::~CFloatLiteral()
{
  delete fString;
}

//
//  Method name : AddToConstantPool
//  Description : This method is used to add this value into the constant
//    pool of the provided class.  It returns the pool index that was used.
//
JavaConstantIndex
CFloatLiteral::AddToConstantPool(CJavaClassFile& inClass) const
{
  JavaConstantIndex index;
  if (GetType() == CJavaTypeSignature::kDouble) {
    index = inClass.AddDoubleConstant(fValue);
  } else {
    index = inClass.AddFloatConstant(fValue);
  }
  return index;
}

//
//  Method name : IsLiteral
//  Description : This method tells the caller whether this expression has a
//    constant value that could be used in a final interface method, a 'case'
//    label, etc.
//
bool
CFloatLiteral::IsLiteral() const
{
  return true;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CFloatLiteral::CreateConstantExpression(CExpression*& result,
					CCompileContext& context) const
{
  result = Clone();
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CFloatLiteral::Unparse(srcstream &sOut,
		       CCompileContext& context,
		       bool isStatement)
{
  sOut.ToLine(fLineNumber,false);
  bool nan = false;
#if defined(HAVE_ISNAN)
  nan = isnan(fValue);
#else
  // By definition, it's supposed to be the case that NaN!=NaN.
  nan = (fValue != fValue);
#endif
  int inf = 0;
  if (!nan) {
#if defined(HAVE_ISINF)
    inf = isinf(fValue);
#elif defined(HAVE_FINITE)
    inf = !finite(fValue);
    if (inf) {
	if (fValue > (double)0.0) {
	    inf = 1;
	} else if (fValue < (double)0.0) {
	    inf = -1;
	} else {
	    inf = 0; nan = true;
	}
    }
#endif
  }
  if (GetType() == CJavaTypeSignature::kDouble) {
    if (nan)
      sOut.out() << " Double.NaN ";
    else if (inf == 1) 
      sOut.out() << " Double.POSITIVE_INFINITY ";
    else if (inf == -1)
      sOut.out() << " Double.NEGATIVE_INFINITY ";
    else if (fString)
      sOut.out() << *fString;
    else
      sOut.out() << setprecision(-1) << fValue; // How portable _is_ this?
  } else {
    assert(GetType() == CJavaTypeSignature::kFloat);
    if (nan)
      sOut.out() << " Float.NaN ";
    else if (inf == 1) 
      sOut.out() << " Float.POSITIVE_INFINITY ";
    else if (inf == -1)
      sOut.out() << " Float.NEGATIVE_INFINITY ";
    else if (fString)
      sOut.out() << *fString;
    else
      sOut.out() << setprecision(-1) << ((float) fValue) << "f";
  }
  return 0;
}

//============================= CStringLiteral ===============================
//
//  Method name : CStringLiteral
//  Description : Constructs a string literal value out of a provided string.
//
CStringLiteral::CStringLiteral(const unicode_string& value)
  : CExpression(CJavaTypeSignature::kString), fValue(value)
{
}

//
//  Method name : CStringLiteral
//  Description : Copy constructor.
//
CStringLiteral::CStringLiteral(const CStringLiteral& source)
  : CExpression(source), fValue(source.fValue)
{
}

//
//  Method name : ~CStringLiteral
//  Description : Destructor
//
CStringLiteral::~CStringLiteral()
{
}

//
//  Method name : AddToConstantPool
//  Description : This method is used to add this value into the constant
//    pool of the provided class.  It returns the pool index that was used.
//
JavaConstantIndex
CStringLiteral::AddToConstantPool(CJavaClassFile& inClass) const
{
  JavaConstantIndex index = inClass.AddAscizConstant(fValue);
  return inClass.AddStringConstant(index);
}

//
//  Method name : IsLiteral
//  Description : This method tells the caller whether this expression has a
//    constant value that could be used in a final interface method, a 'case'
//    label, etc.
//
bool
CStringLiteral::IsLiteral() const
{
  return true;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CStringLiteral::CreateConstantExpression(CExpression*& result,
					CCompileContext& context) const
{
  result = Clone();
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CStringLiteral::Unparse(srcstream &sOut,
			CCompileContext& context,
			bool isStatement)
{
  sOut.ToLine(fLineNumber,false).out() << '\"' 
				       << ::StrLitToString(fValue)
				       << '\"';
  return 0;
}


//============================== CNull =================================

//
//  Method name : CNull
//  Description : Constructs a null literal value.
//
CNull::CNull()
  : CExpression(CJavaTypeSignature::kNullType)
{
}

//
//  Method name : CNull
//  Description : copy constructor
//
CNull::CNull(const CNull& other)
  : CExpression(other)
{
}

//
//  Method name : ~CNull
//  Description : Destructor
//
CNull::~CNull()
{
}


//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CNull::CreateConstantExpression(CExpression*& result,
				CCompileContext& context) const
{
  result = Clone();
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CNull::Unparse(srcstream &sOut,
	       CCompileContext& context,
	       bool isStatement)
{
  sOut.ToLine(fLineNumber,false).out() << "null";
  return 0;
}

//============================= CCastExpression ===============================

//
//  Method name : CCastExpression
//  Description : Constructs an expression casting a subexpression to another
//    type.
//
CCastExpression::CCastExpression(const CJavaTypeSignature& toType,
				 CExpression* adoptCastee,
				 bool fast)
  : CExpression(CJavaTypeSignature::kVoid),
    fCastToType(toType),
    fCastee(adoptCastee),
    fFastCast(fast)
{
  fPrecedence = 1;
  fLineNumber = fCastee->GetLineNumber();
}

//
//  Method name : CCastExpression
//  Description : Copy constructor.
//
CCastExpression::CCastExpression(const CCastExpression& source)
  : CExpression(source),
    fCastToType(source.fCastToType),
    fCastee(0),
    fFastCast(source.fFastCast)   
{
  if (source.fCastee != 0) {
    fCastee = source.fCastee->Clone();
  }
}

//
//  Method name : ~CCastExpression
//  Description : Destructor
//
CCastExpression::~CCastExpression()
{
  delete fCastee;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CCastExpression::HandleEvaluateType(CCompileContext& context,
				    CJavaTypeSignature& setType)
{
  assert(fCastee != 0);
  CCompileError* error = context.GetCompiler().CheckType(fCastToType,
							 context.GetClass(),
							 context.GetSource());
  if (error) { 
    error->SetLine(fLineNumber);
    return error;
  }
  setType = context.GetCompiler().FixType(fCastToType,
					  context.GetSource());
  fCastToType = setType;
  CJavaTypeSignature type;
  error = EvaluateType(fCastee, context, type);
  if (error == 0 &&
      !context.GetCompiler().SameType(type, fCastToType,
				      context.GetSource()) &&
      !context.GetCompiler().AssignableSubtype(type, fCastToType,
					       context.GetSource())) {
    context.GetCompiler().WarnUnenforcedCast(fCastToType,
					     context, GetLineNumber());
    // Are we casting to a parameter type?
    if (fCastToType.IsParameter()) {
      unicode_string className;
      if (type.GetBaseClassName(className) && 
	  className == CCompiler::kObjectName) {
	return 0;
      }
    }
    if (type.IsReference() && fCastToType.IsReference()) {
      if (!context.GetCompiler().CastableType(type, fCastToType,
					      context.GetSource())) {
	error = MakeBadCastMessage(type);
      }
    } else if (!type.IsNumeric() || !fCastToType.IsNumeric()) {
      error = MakeBadCastMessage(type);
    }
  }      
  return error;
}

//
//  Method name : MakeBadCastMessage
//  Description : An internal method used by CCastExpression to generate a
//    compile error for an invalid conversion from 'castFrom' type to this
//    type.
//
CCompileError*
CCastExpression::MakeBadCastMessage(const CJavaTypeSignature& castFrom) const
{
  unicode_string errorString = ::StringToUnicode("Cannot convert ");
  errorString += castFrom.Disassemble();
  errorString += ::StringToUnicode(" value to ");
  errorString += fCastToType.Disassemble();
  return new CCompileError(errorString, fLineNumber);
}

//
//  Method name : CastPointer
//  Description : This static method is provided so that intermediate stages
//    of compilation can change the type of an expression pointer that they
//    are holding by casting it to a different value.  This performs the
//    cast, updates the new expression, and then checks the validity of the
//    resulting typecast.  If there is a problem, it returns an error,
//    otherwise it returns 0.
//  * Possible bug: CastPointer always returns a "fast" cast.
//
CCompileError*
CCastExpression::CastPointer(CExpression*& expressionPointer,
			     const CJavaTypeSignature& toType,
			     CCompileContext& context)
{
  assert(expressionPointer != 0);
  CExpression* newExpression = new CCastExpression(toType, 
						   expressionPointer, true);
  newExpression->InheritLineNumber(*expressionPointer);
  expressionPointer = newExpression;
  CJavaTypeSignature temporaryType;
  return EvaluateType(expressionPointer, context, temporaryType);
}

//
//  Method name : ImplicitCastPointer
//  Description : This static function attempts to convert the provided
//    expression into the provided type via an implicit type cast. If an
//    implicit cast isn't possible between these types, an error is returned,
//    otherwise null is returned.
//  * Possible bug: ImplicitCastPointer always returns a "fast" cast.
//
CCompileError*
CCastExpression::ImplicitCastPointer(CExpression*& expressionPointer,
				     const CJavaTypeSignature& toType,
				     CCompileContext& context)
{
  CCompileError* error = 0;
  CJavaTypeSignature castFrom = expressionPointer->GetType();
  if (context.GetCompiler().ImplicitCastTo(castFrom, toType,
					   context.GetSource())) {
    error = CastPointer(expressionPointer, toType, context);
  } else {
    COrdinalLiteral* literal =
      DYNAMIC_CAST(COrdinalLiteral, expressionPointer);
    if (literal != 0 &&
	(toType == CJavaTypeSignature::kShort ||
	 toType == CJavaTypeSignature::kCharacter ||
	 toType == CJavaTypeSignature::kByte)) {
      u64 value = literal->GetLong();
      bool converted = false;
      switch (toType.GetBaseType()) {
      case CJavaTypeSignature::Short:
      case CJavaTypeSignature::Character:
	{
	  short shortValue = (short)value;
	  if ((u64) shortValue == value) {
	    delete expressionPointer;
	    expressionPointer = new COrdinalLiteral(shortValue,
				    toType == CJavaTypeSignature::kCharacter);
	    converted = true;
	  }
	}
	break;
      case CJavaTypeSignature::Byte:
	{
	  signed char byteValue = (signed char)value;
	  if ((u64) byteValue == value) {
	    delete expressionPointer;
	    expressionPointer = new COrdinalLiteral(byteValue);
	    converted = true;
	  }	
	}  
	break;
      default:
	assert(0);
      }
      if (!converted) {
	unicode_string errorString =
	  ::StringToUnicode("Cannot convert constant integer value to ");
	errorString += toType.Disassemble();
	errorString += ::StringToUnicode(" without an explicit cast.");
	error =
	  new CCompileError(errorString, expressionPointer->GetLineNumber());
      }
    } else {
      unicode_string errorString = ::StringToUnicode("Cannot convert ");
      errorString += castFrom.Disassemble();
      errorString += ::StringToUnicode(" value to ");
      errorString += toType.Disassemble();
      errorString += ::StringToUnicode(" via an implicit cast.");
      error =
	new CCompileError(errorString, expressionPointer->GetLineNumber());
    }
  }
  return error;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CCastExpression::CreateConstantExpression(CExpression*& result,
					  CCompileContext& context) const
{
  assert(fCastee != 0);
  if (fCastee->IsLiteral()) {
    if (fCastToType.IsNumeric()) {
      COrdinalLiteral* intLiteral = DYNAMIC_CAST(COrdinalLiteral, fCastee);
      if (intLiteral != 0) {
	switch(fCastToType.GetBaseType()) {
	case CJavaTypeSignature::Integer:
	  result = new COrdinalLiteral((JavaSInt)intLiteral->GetInteger());
	  break;
	case CJavaTypeSignature::LongInteger:
	  result = new COrdinalLiteral((JavaSLong)intLiteral->GetLong());
	  break;
	case CJavaTypeSignature::Byte:
	  result = new COrdinalLiteral((JavaUByte)intLiteral->GetByte());
	  break;
	case CJavaTypeSignature::Short:
	  result = new COrdinalLiteral((JavaSShort)intLiteral->GetShort());
	  break;
	case CJavaTypeSignature::Character:
	  result = new COrdinalLiteral((JavaUChar)intLiteral->GetShort(),
				       true);
	  break;
	case CJavaTypeSignature::Float:
	  result = new CFloatLiteral((float)(long long)intLiteral->GetLong());
	  break;
	case CJavaTypeSignature::Double:
	   result =
	     new CFloatLiteral((double)(long long)intLiteral->GetLong());
	  break;
	default:
	  assert(0);
	}
      } else {
	CFloatLiteral* floatLiteral = DYNAMIC_CAST(CFloatLiteral, fCastee);
	if (floatLiteral != 0) {
	  switch(fCastToType.GetBaseType()) {
	  case CJavaTypeSignature::Integer:
	    result =
	      new COrdinalLiteral((JavaUInt)floatLiteral->GetDouble());
	    break;
	  case CJavaTypeSignature::LongInteger:
	    result = new COrdinalLiteral(
			   (JavaULong)floatLiteral->GetDouble());
	    break;
	  case CJavaTypeSignature::Byte:
	    result =
	      new COrdinalLiteral((JavaUByte)floatLiteral->GetDouble());
	    break;
	  case CJavaTypeSignature::Short:
	    result =
	      new COrdinalLiteral((JavaUShort)floatLiteral->GetDouble(),
				  false);
	    break;
	  case CJavaTypeSignature::Character:
	    result =
	      new COrdinalLiteral((JavaUChar)floatLiteral->GetDouble());
	    break;
	  case CJavaTypeSignature::Float:
	    result = new CFloatLiteral(floatLiteral->GetFloat());
	    break;
	  case CJavaTypeSignature::Double:
	    result = new CFloatLiteral(floatLiteral->GetDouble());
	    break;
	  default:
	    assert(0);
	  }
	}
      }
    }
  }
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CCastExpression::Unparse(srcstream &sOut,
			 CCompileContext& context,
			 bool isStatement)
{
  assert(fCastee != 0);
  CJavaTypeSignature type = fCastee->GetType();
  CCompileError* error = 
    context.GetCompiler().CheckType(fCastToType,context.GetClass(),
				    context.GetSource());
  if (error) {
    error->SetLine(fLineNumber);  
    return error;
  }

  sOut.ToLine(fLineNumber,false);
  if (fFastCast || 
      (!fCastToType.IsParameter() && 
       (!fCastToType.IsParameterized() ||
	fCastToType.IsArray()))) {
    // Just cast it.
    sOut.out() << "((";
    fCastToType.Unparse(sOut); 
    sOut.out() << ") (";
    error = fCastee->Unparse(sOut,context);
    sOut.out() << "))";
  } else if (fCastToType.IsParameter()) { 
    // Use a Cast auxmethod.    
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::Cast, fCastToType);
    if (!isStatement) sOut.out() << "(";
    sOut.out() << "("; 
    fCastToType.Unparse(sOut);
    sOut.out() << ") $W."
	       << ::UnicodeToUTF(auxMethod)
	       << "((Object) ";
    error = fCastee->Unparse(sOut,context);
    sOut.out() << (isStatement ?  ")" : "))" );
  } else {
    // Call a $cast method.
    assert(fCastToType.IsParameterized() && ! fCastToType.IsParameter());
    fCastToType.UnparseWType(sOut);
    sOut.out() << ".$cast(";
    error = fCastee->Unparse(sOut,context);
    for (deque<CJavaTypeSignature>::const_iterator i = 
	   fCastToType.GetParameters()->begin();
	 i != fCastToType.GetParameters()->end(); ++i) {
      if ((*i).IsInstantiable()) {
	sOut.out() << ", \"" << (*i).Compile() << "\"";
      } else {
	unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
	  AuxMethodName(CAuxMethodSignature::PSig, (*i));
	sOut.out() << ", $W." << ::UnicodeToUTF(auxMethod) << "()";
      }
    }
    sOut.out() << ")";
    if (fCastToType.IsInstantiable()) {
      CJavaTypeSignature t = context.GetCompiler().FixType(fCastToType,
						        context.GetSource());
      context.GetCompiler().GetIList().AddType(t, context.GetSource());
    }
  } 
  return error;
}

//============================= CNewArray ================================

//
//  Method name : CNewArray
//  Description : Constructs an allocation expression for an array of objects.
//
CNewArray::CNewArray(const CJavaTypeSignature& type,
		     ExpressionList* adoptArraySizes)
  : CExpression(type),
    fArraySizes(adoptArraySizes),
    fBase(0),
    fOptDims(0)
{
  fPrecedence = 100; 
  assert(fArraySizes!=0 && !fArraySizes->empty() && fArraySizes->front()!=0);

  fLineNumber = fArraySizes->front()->GetLineNumber();
}

//
//  Method name : CNewArray
//  Description : Constructor for the ambiguous array creation case.
CNewArray::CNewArray(CExpression* adoptBase, 
		     ExpressionList* adoptArraySizes,
		     int optDims)
  : CExpression(CJavaTypeSignature::Void),
    fArraySizes(adoptArraySizes),
    fBase(adoptBase),
    fOptDims(optDims)
{
}

//
//  Method name : CNewArray
//  Description : Copy constructor.
//
CNewArray::CNewArray(const CNewArray& source)
  : CExpression(source),
    fArraySizes(0),
    fBase(source.fBase->Clone()),
    fOptDims(source.fOptDims)
{
  if (source.fArraySizes != 0) {
    fArraySizes = new ExpressionList;
    for (ExpressionList::iterator i = source.fArraySizes->begin();
	 !(i == source.fArraySizes->end()); ++i) {
      fArraySizes->push_back((*i)->Clone());
    }
  }
}

//
//  Method name : ~CNewArray
//  Description : Destructor.
//
CNewArray::~CNewArray()
{
  for (ExpressionList::iterator i = fArraySizes->begin();
       !(i == fArraySizes->end()); ++i) {
    delete *i;
  }
  delete fArraySizes;
  delete fBase;
}

//
//  Method name : Resolve
//  Description : Attempts to resolve any ambiguities in this expression.
//      Does not attempt to resolve ambiguities in subexpressions.  On
//      failure, returns an error message.
//
CCompileError* 
CNewArray::Resolve(CCompileContext& context,
		   CExpression*& setExpression)
{
  if (fBase) {
    CAmbiguousAccess* base = DYNAMIC_CAST(CAmbiguousAccess, fBase);
    assert(base);
    CJavaTypeSignature* type = base->ToType();
    if (type) {
      const CJavaClassFile* classOfType = 
	context.GetCompiler().ClassOfType(*type,context.GetSource());
      if (classOfType && classOfType->IsParameterized()) {
	type->SetArrayBounds(fOptDims + fArraySizes->size());
	SetType(*type);
	delete fBase;
	delete type;
	fBase = 0;
	return 0;
      }
    }
    // Oops -- fBase _does_ contain an index!
    CExpression *exp = 0;
    CCompileError *err = base->ToTypeAndDimension(fType, exp);
    if (err) return err;
    fArraySizes->push_front(exp);
    fType.SetArrayBounds(fOptDims + fArraySizes->size());
    delete fBase;
    fBase = 0;
  }
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CNewArray::Unparse(srcstream &sOut,
		   CCompileContext& context,
		   bool isStatement)
{
  CCompileError* error = 
    context.GetCompiler().CheckType(fType,
				    context.GetClass(),
				    context.GetSource());
  if (error) {
    error->SetLine(fLineNumber);
    return error;
  }
  if ((fArraySizes == 0) || ! GetType().IsArray()) {
    return MakeError("Bad array constructor call");
  }
  if (GetType().IsParameter()) {
    CJavaTypeSignature type = context.GetCompiler().FixType(GetType(),
							 context.GetSource());
    type.SetArrayBounds(0);
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::ArrayNew, type,
		    0,0,0,CAuxMethodSignature::Virtual,
		    (unsigned) (fArraySizes)->size(),
		    GetType().GetArrayBounds());
    sOut.ToLine(fLineNumber).out()
      << "$W."
      << ::UnicodeToUTF(auxMethod)
      << '(';
    for (ExpressionList::iterator sizeIterator = fArraySizes->begin();
	 error == 0 && !(sizeIterator == fArraySizes->end()); ) {
      CJavaTypeSignature arraySizeType;
      assert(*sizeIterator != 0);
      error = EvaluateType((*sizeIterator), context, arraySizeType);
      if (error == 0) {
	if (!UnaryPromote(*sizeIterator, context)) {
	  error = MakeError("Cannot convert array size to integer type.");
	} else {
	  error = (*sizeIterator)->Unparse(sOut,context);
	}
      }
      ++sizeIterator;
      if (sizeIterator != fArraySizes->end())
	sOut.out() << ", ";
    }
    sOut.out() << ")";
    return error;
  }
  // No, it's not a parameter.

  sOut.ToLine(fLineNumber).out() << "new ";
  fType.UnparseBaseType(sOut);
  for (ExpressionList::iterator sizeIterator = fArraySizes->begin();
       error == 0 && !(sizeIterator == fArraySizes->end()); ++sizeIterator) {
    CJavaTypeSignature arraySizeType;
    assert(*sizeIterator != 0);
    error = EvaluateType((*sizeIterator), context, arraySizeType);
    if (error == 0) {
      if (!UnaryPromote(*sizeIterator, context)) {
	error = MakeError("Cannot convert array size to integer type.");
      } else {
	sOut.out() << '[';
	error = (*sizeIterator)->Unparse(sOut,context);
	sOut.out() << ']';
      }
    }
  }
  for (unsigned long i = 0; 
       i < (GetType().GetArrayBounds() - fArraySizes->size()); ++i) {
    sOut.out() << "[]";
  }

  return error;
}

//============================= CNewObject ================================

//
//  Method name : CNewObject
//  Description : Constructs an allocation expression to allocate a class
//    value.  The first parameter to this function is the name of the class
//    to allocate and the second is a list of arguments to be given to the
//    class constructor at creation time.
//
CNewObject::CNewObject(const unicode_string& className,
		       ExpressionList* adoptArgs)
  : CExpression(className),
    fConstructorArguments(adoptArgs)
{
  fPrecedence = 100;
  if (fConstructorArguments == 0) {
    fConstructorArguments = new ExpressionList;
  }
  if (!fConstructorArguments->empty()) 
    fLineNumber = fConstructorArguments->front()->GetLineNumber();
}

//
//  Method name : CNewObject
//  Description : Constructs an allocation expression to allocate a class
//    value.  The first parameter to this function is the type of the class
//    to allocate and the second is a list of arguments to be given to the
//    class constructor at creation time.
//
CNewObject::CNewObject(const CJavaTypeSignature& type,
		       ExpressionList* adoptArgs)
  : CExpression(type),
    fConstructorArguments(adoptArgs)
{
  fPrecedence = 100;
  if (fConstructorArguments == 0) {
    fConstructorArguments = new ExpressionList;
  }
  if (!fConstructorArguments->empty()) 
    fLineNumber = fConstructorArguments->front()->GetLineNumber();
}

//
//  Method name : CNewObject
//  Description : Copy constructor.
//
CNewObject::CNewObject(const CNewObject& source)
  : CExpression(source),
    fConstructorArguments(0)
{
  if (source.fConstructorArguments != 0) {
    fConstructorArguments = new ExpressionList;
    for (ExpressionList::iterator i = source.fConstructorArguments->begin();
	 !(i == source.fConstructorArguments->end()); ++i) {
      fConstructorArguments->push_back((*i)->Clone());
    }
  }
}

//
//  Method name : ~CNewObject
//  Description : Destructor.
//
CNewObject::~CNewObject()
{
  delete fConstructorArguments;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CNewObject::Unparse(srcstream &sOut,
		    CCompileContext& context,
		    bool isStatement)
{
  CCompileError* error = 0;
  unicode_string className;
  bool isClass = GetType().GetBaseClassName(className);
  assert(isClass && !GetType().IsArray());
  const CJavaClassFile* javaClass = 0;
  int methodidx = 0;
  bool needW = false;
  if (!GetType().IsParameter()) {
    // Not a parameter: make sure we can lookup the class.
    javaClass = context.GetCompiler().LookupClass(className,
						  context.GetSource());
    if (javaClass == 0) {
      string errorString("Cannot create an object of unknown type ");
      errorString += ::SymbolToString(className);
      error = MakeError(errorString);
    } else if (javaClass->GetAccessFlags().fAbstract) {
      string errorString("Cannot instantiate abstract class ");
      errorString += ::SymbolToString(javaClass->GetClassName());
      error = MakeError(errorString);
    } else if (javaClass->GetAccessFlags().fInterface) {
      string errorString("Cannot instantiate interface ");
      errorString += ::SymbolToString(javaClass->GetClassName());
      error = MakeError(errorString);
    } else {
      error = context.GetCompiler().CheckType(GetType(),context.GetClass(),
					      context.GetSource());
      if (error) error->SetLine(fLineNumber);  
    }
  }
  if (error) return error;
  
  // Find the types of the arguments.
  deque<CJavaTypeSignature> argumentTypes;  
  for (ExpressionList::iterator i = fConstructorArguments->begin();
       error == 0 && i != fConstructorArguments->end(); ++i) {
    CJavaTypeSignature argumentType;
    error = EvaluateType(*i, context, argumentType);
    argumentTypes.push_back(argumentType);
  }
  if (error) return error;

  CJavaMethodSignature methodSignature;
  CJavaMethodSignature transMethodSignature;
  if (!GetType().IsParameter()) {
    ////
    // We don't have a parameter: lookup the constructor
    ////
    string* matchErrorString = 0;
    CJavaTypeSignature resType;
    pair<const CJavaClassFile*, const CJavaMethodInfo*>* match =
      context.GetCompiler().MatchMethod(CCompiler::kConstructorName,
					GetType(), 
					argumentTypes, matchErrorString,
					resType, context.GetSource());
    if (match == 0 || match->first != javaClass) {
      // We didn't find the constructor.
      if (matchErrorString == 0) {
	string errorString;
	errorString = "No matching constructor for class ";
	errorString += ::SymbolToString(className);
	errorString += '(';
	deque<CJavaTypeSignature>::iterator arg = argumentTypes.begin();
	while (arg != argumentTypes.end()) {
	  errorString += ::UnicodeToString((*arg).Disassemble());
	  if (++arg != argumentTypes.end()) {
	    errorString += ", ";
	  }
	}
	errorString += ')';
	error = MakeError(errorString);
      } else {
	error = MakeError(*matchErrorString);
	delete matchErrorString;
      }
    } else {
      // We found the constructor.

      // Check the throws.
      for (deque<unicode_string>::const_iterator i = 
	     match->second->ThrowsBegin();
	   error == 0 && i != match->second->ThrowsEnd(); ++i) {
	if (!context.Throwable(CJavaTypeSignature(*i))) {
	  string message = ::SymbolToString(*i);
	  message += " must be caught or declared in this "
	    "method's 'throws' list.";
	  error = MakeError(message);
	}
      }
      if (error) return error;

      sOut.ToLine(fLineNumber);
      CInstantiator ins(*match->first, resType.GetParameters());
      transMethodSignature = match->second->GetTranslatedSignature();
      ins.Instantiate(methodSignature, match->second->GetSignature());
      
      sOut.out() << "new ";
      GetType().Unparse(sOut);
      sOut.out() << '(';

      // Do we need a where object?
      if (GetType().IsParameterized()) {
	needW = true;
	UnparseWClassCode(GetType(), context, sOut);
	if (fConstructorArguments->size() > 0) sOut.out() << ',';
      }	
      if (javaClass->HasPInfo()) {
	// Do we need to mangle the name?
	assert(javaClass->GetPInfo().GetMethods());
	MethodNameMap::const_iterator found = javaClass->GetPInfo().
	  GetMethods()->find(match->second->GetSignature());
	if (found != javaClass->GetPInfo().GetMethods()->end()) {
	  methodidx = (*found).second.fIndex;
	}
      }
      delete match;
    }
  } else {
    ////
    // Okay; it _is_ a parameter.  Call the proper auxMethod.
    ////
    CJavaMethodSignature m(CJavaTypeSignature::kVoid,
			   CCompiler::kConstructorName, argumentTypes);
    const CPJavaConstraint* cons = 
      context.GetClass().GetPInfo().LookupConstraint(className);
    assert(cons);
    const CPJavaConstraint::ConstraintMethod* match = 
      context.GetCompiler().MatchConstraintMethod(m, *cons);
    if (match == 0) {
      // We didn't find the constructor.
      string errorString;
      errorString = "No matching constructor for parameter ";
      errorString += ::SymbolToString(className);
      errorString += '(';
      deque<CJavaTypeSignature>::iterator arg = argumentTypes.begin();
      while (arg != argumentTypes.end()) {
	errorString += ::UnicodeToString((*arg).Disassemble());
	if (++arg != argumentTypes.end()) {
	  errorString += ", ";
	}
      }
      errorString += ')';
      error = MakeError(errorString);
    } else {
      // We found the right constructor.
      
      // Check the throws.
      for (deque<CJavaTypeSignature>::iterator i = match->fThrows->begin(); 
	   error == 0 && i != match->fThrows->end(); ++i) {
	if (!context.Throwable(*i)) {
	  unicode_string message = (*i).Disassemble();
	  message += ::UTFToUnicode(" must be caught or declared in this "
				    "method's 'throws' list.");
	  error = MakeError(message);	
	}
      }
      if (error) return error;

      methodSignature = *match->fMethod;
      transMethodSignature = *match->fMethod;
      transMethodSignature.Objectify();
      unicode_string methodName = match->fMethod->GetName();
      unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
	AuxMethodName(CAuxMethodSignature::Where,
		      GetType(),
		      &methodName,
		      (match->fMethod),
		      (match->fThrows),
		      CAuxMethodSignature::Constructor);
      sOut.ToLine(fLineNumber).out() 
	<< "$W." << ::UnicodeToUTF(auxMethod) << '(';
    }
  }
  if (error) return error;

  // Now write out the Arguments
  deque<CJavaTypeSignature>::const_iterator parameter =
    methodSignature.ParametersBegin();
  deque<CJavaTypeSignature>::const_iterator transParameter =
    transMethodSignature.ParametersBegin();
  if (transParameter != transMethodSignature.ParametersEnd()
      && ((*transParameter).IsWhereType() && 
	  (parameter == methodSignature.ParametersEnd() || 
	   !(*parameter).IsWhereType())))
    ++transParameter;
  for (ExpressionList::iterator i = fConstructorArguments->begin();
       error == 0 && i != fConstructorArguments->end(); ) {
    if (!error)
      error = CExpression::UnparseArgument(*parameter,
					   *transParameter,
					   *(*i),
					   sOut, context);
    ++i;
    if (i != fConstructorArguments->end())
      sOut.out() << ", ";
    ++parameter;
    ++transParameter;
  }  
  // Add dummy arguments
  if (methodidx > 0 && ((fConstructorArguments->size() > 0)
			|| needW))
    sOut.out() << ", ";
  for (int i = 0; i < methodidx; ++i) {
    sOut.out() << "0";
    if (i < methodidx-1) sOut.out() << ", ";
  }
  sOut.out() << ")";
  
  return error;
}


//=========================== CTrinaryExpression ==============================

//
//  Method name : CTrinaryExpression
//  Description : Constructs a Trinary expression from the three child
//    expressions that make it up.
//
CTrinaryExpression::CTrinaryExpression(CExpression* leftExpression,
				       CExpression* middleExpression,
				       CExpression* rightExpression)
  : CExpression(CJavaTypeSignature::Void, false),
    fLeft(leftExpression),
    fMiddle(middleExpression),
    fRight(rightExpression)
{
  fPrecedence = 13;
  fLineNumber = fLeft->GetLineNumber();
}

//
//  Method name : CTrinaryExpression
//  Description : Copy constructor.
//
CTrinaryExpression::CTrinaryExpression(const CTrinaryExpression& source)
  : CExpression(source), fLeft(0), fMiddle(0), fRight(0)
{
  if (source.fLeft != 0) {
    fLeft = source.fLeft->Clone();
  }
  if (source.fMiddle != 0) {
    fMiddle = source.fMiddle->Clone();
  }
  if (source.fRight != 0) {
    fRight = source.fRight->Clone();
  }
}

//
//  Method name : ~CTrinaryExpression
//  Description : Destructor.
//
CTrinaryExpression::~CTrinaryExpression()
{
  delete fLeft;
  delete fMiddle;
  delete fRight;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CTrinaryExpression::HandleEvaluateType(CCompileContext& context,
				       CJavaTypeSignature& setType)
{
  assert(fLeft != 0 && fMiddle != 0 && fRight != 0);
  CJavaTypeSignature leftType, middleType, rightType;
  CCompileError* error = EvaluateType(fLeft, context, leftType);
  if (error == 0) {
    error = EvaluateType(fMiddle, context, middleType);
  }
  if (error == 0) {
    error = EvaluateType(fRight, context, rightType);
  }
  if (error == 0 && leftType != CJavaTypeSignature::kBoolean) {
    error = MakeError("Non-boolean expression used in conditional expression");
  }
  if (error == 0 && (rightType == CJavaTypeSignature::kVoid ||
		     middleType == CJavaTypeSignature::kVoid)) {
    error = MakeError("Invalid use of void-typed expression");
  }
  if (error == 0) {
    if (context.GetCompiler().SameType(middleType, 
				       rightType, context.GetSource())) {
      setType = middleType;
    } else if (middleType.IsNumeric() && rightType.IsNumeric()) {
      if (!UnifyNumeric(fMiddle, fRight, context, setType) &&
	  !UnifyNumeric(fRight, fMiddle, context, setType)) {
	if (BinaryPromote(fMiddle, fRight, context)) {
	  setType = fRight->GetType();
	} else {
	  error = MakeError("Incompatible types in conditional expression");
	}
      }
    } else if (middleType.IsReference() && rightType.IsReference()) {
      if (middleType == CJavaTypeSignature::kNullType) {
	error = CCastExpression::CastPointer(fMiddle, rightType, context);
	setType = rightType;
      } else if (rightType == CJavaTypeSignature::kNullType) {
	error = CCastExpression::CastPointer(fRight, middleType, context);
	setType = middleType;
      } else if (context.GetCompiler().AssignableSubtype(middleType,
					    rightType,context.GetSource())) {
	error = CCastExpression::CastPointer(fMiddle, rightType, context);
	setType = rightType;
      } else if (context.GetCompiler().AssignableSubtype(rightType,
					  middleType,context.GetSource())) {
	error = CCastExpression::CastPointer(fRight, middleType, context);
	setType = middleType;
      } else {
	error = MakeError("Incompatible types in conditional expression");
      }
    } else {
      error = MakeError("Incompatible types used in conditional expression");
    }
  }
  return error;
}

//
//  Method name : UnifyNumeric
//  Description : This method is used internally to try to unify two numeric
//    values to this trinary expression according to the rules in section
//    15.24 of the language specification.  If this unification is successful,
//    this method will return true and set the value of 'setType' accordingly.
//    Otherwise, this returns false.
//    This method assumes that the types of these expressions has already
//    been successfully evaluated.
//
bool
CTrinaryExpression::UnifyNumeric(CExpression*& first, CExpression*& second,
		       CCompileContext& context, CJavaTypeSignature& setType)
{
  COrdinalLiteral* literal = DYNAMIC_CAST(COrdinalLiteral, second);
  u64 value;
  if (literal != 0) {
    value = literal->GetLong();
  }
  switch (first->GetType().GetBaseType()) {
  case CJavaTypeSignature::Short:
    if (second->GetType() == CJavaTypeSignature::kByte) {
      setType = CJavaTypeSignature::kShort;
      CCompileError* error =
	CCastExpression::CastPointer(second, setType, context);
      // casting a Byte value to a Short should always work.
      assert(error == 0);
      return true;
    } 
  case CJavaTypeSignature::Character:
    if (literal != 0) {
      short shortValue = (short)value;
      setType = first->GetType();
      if ((u64) shortValue == value) {
	delete second;
	second = new COrdinalLiteral(shortValue,
				    setType == CJavaTypeSignature::kCharacter);
	return true;
      }
    }
    break;
  case CJavaTypeSignature::Byte:
    if (literal != 0) {
      setType = first->GetType();
      signed char byteValue = (signed char)value;
      if ((u64) byteValue == value) {
	delete second;
	second = new COrdinalLiteral(byteValue);
	return true;
      }
    }
    break;
  default:
    break;
  }
  return false;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CTrinaryExpression::CreateConstantExpression(CExpression*& result,
					     CCompileContext& context) const
{
  result = 0;
  if (fLeft->IsLiteral()) {
    COrdinalLiteral* left = DYNAMIC_CAST(COrdinalLiteral, fLeft);
    assert(left != 0);
    CExpression* selected = left->GetBoolean() ? fMiddle : fRight;
    if (selected->IsLiteral()) {
      result = selected->Clone();
    }
  }
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CTrinaryExpression::Unparse(srcstream &sOut,
			    CCompileContext& context,
			    bool isStatement)
{
  CCompileError* error = 0;
  assert(fLeft != 0 && fMiddle != 0 && fRight != 0);
  bool parenLeft   = (fPrecedence <=   fLeft->GetPrecedence());
  bool parenRight  = (fPrecedence <=  fRight->GetPrecedence());
  bool parenMiddle = (fPrecedence <= fMiddle->GetPrecedence());
  if (parenLeft)
    sOut.ToLine(fLineNumber,false).out() << '(';
  error = fLeft->Unparse(sOut,context);
  if (parenLeft)
    sOut.out() << ')';
  sOut.out() << " ? ";
  if (parenMiddle)
    sOut.out() << '(';
  if (error == 0) 
    error = fMiddle->Unparse(sOut,context);
  if (parenMiddle)
    sOut.out() << ')';
  sOut.out() << " : ";
  if (parenRight)
    sOut.out() << '(';
  if (error == 0) 
    error = fRight->Unparse(sOut,context);
  if (parenRight)
    sOut.out() << ')';

  return error;
}

//=========================== CUnaryExpression ================================

//
//  Method name : CUnaryExpression
//  Description : Constructs a unary expression out of the constituent parts:
//    the child expression, an operator to apply to it, and a flag to tell
//    whether it is an lvalue or not.
//
CUnaryExpression::CUnaryExpression(CExpression* childExpression,
		     unsigned short unaryOperator, Order side, bool isLValue)
  : CExpression(CJavaTypeSignature::Void, isLValue),
    fChild(childExpression),
    fOperator(unaryOperator),
    fSide(side)
{
  fPrecedence = 1;
}

//
//  Method name : CUnaryExpression
//  Description : Copy constructor.
//
CUnaryExpression::CUnaryExpression(const CUnaryExpression& source)
  : CExpression(source),
    fChild(0),
    fOperator(source.fOperator),
    fSide(source.fSide)
{
  if (source.fChild != 0) {
    fChild = source.fChild->Clone();
  }
}

//
//  Method name : ~CUnaryExpression
//  Description : Destructor
//
CUnaryExpression::~CUnaryExpression()
{
  delete fChild;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CUnaryExpression::HandleEvaluateType(CCompileContext& context,
				     CJavaTypeSignature& setType)
{
  assert(fChild != 0);
  CJavaTypeSignature type;
  CCompileError* error = EvaluateType(fChild, context, type,
				      fOperator == INCR || fOperator == DECR);
  if (error == 0) {
    switch (fOperator) {
    case '-':
    case '+':
      if (!type.IsNumeric()) {
	error = MakeError("Non-numeric value used in unary expression");
      } else {
	UnaryPromote(fChild, context);
	setType = fChild->GetType();
      }
      break;
    case INCR:
    case DECR:
      if (!type.IsNumeric()) {
	error = MakeError("Non-numeric value used in unary expression");
      } else if (fChild->IsLValueAlias()) {
	string msg = "Cannot use ";
	msg += (fOperator == INCR) ? "++" : "--";
	msg += DYNAMIC_CAST(CClassFieldExpression, fChild) ?
	  " on static field of PClass" :
	  " on array of parameter type";
	error = MakeError(msg);
      } else {
	setType = type;
      }
      break;
    case '~':
      if (!type.IsOrdinal()) {
	error = MakeError("Non-integral value used with unary '~' expression");
      } else {
	UnaryPromote(fChild, context);
	setType = fChild->GetType();
      }
      break;
    case '!':
      if (type != CJavaTypeSignature::kBoolean) {
	error = MakeError("Non-boolean value used with '!' expressions");
      } else {
	setType = type;
      }
      break;
    default:
      assert(0);
    }
  }
  return error;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CUnaryExpression::CreateConstantExpression(CExpression*& result,
					     CCompileContext& context) const
{
  result = 0;
  assert(fChild != 0);
  if (fChild->IsLiteral()) {
    COrdinalLiteral* intLiteral = DYNAMIC_CAST(COrdinalLiteral, fChild);
    CFloatLiteral* floatLiteral = DYNAMIC_CAST(CFloatLiteral, fChild);
    switch (fOperator) {
    case '-':
      switch (fChild->GetType().GetBaseType()) {
      case CJavaTypeSignature::LongInteger:
	assert(intLiteral != 0);
	result = new COrdinalLiteral(-(JavaSLong)intLiteral->GetLong());
	break;
      case CJavaTypeSignature::Integer:
	assert(intLiteral != 0);
	result = new COrdinalLiteral(-(JavaSInt)intLiteral->GetInteger());
	break;
      case CJavaTypeSignature::Float:
	assert(floatLiteral != 0);
	result = new CFloatLiteral(-floatLiteral->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	assert(floatLiteral != 0);
	result = new CFloatLiteral(-floatLiteral->GetDouble());
	break;
      default:
	assert(0);
      }
      break;
    case '+':
      switch (fChild->GetType().GetBaseType()) {
      case CJavaTypeSignature::LongInteger:
	assert(intLiteral != 0);
	result = new COrdinalLiteral(+intLiteral->GetLong());
	break;
      case CJavaTypeSignature::Integer:
	assert(intLiteral != 0);
	result = new COrdinalLiteral(+intLiteral->GetInteger());
	break;
      case CJavaTypeSignature::Float:
	assert(floatLiteral != 0);
	result = new CFloatLiteral(+floatLiteral->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	assert(floatLiteral != 0);
	result = new CFloatLiteral(+floatLiteral->GetDouble());
	break;
      default:
	assert(0);
      }
      break;
    case '~':
      assert(intLiteral != 0);
      switch (intLiteral->GetType().GetBaseType()) {
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral(~intLiteral->GetLong());
	break;
      default:
	result = new COrdinalLiteral(~intLiteral->GetInteger());
      }
      break;
    case '!':
      assert(intLiteral != 0);
      result = new COrdinalLiteral(!intLiteral->GetBoolean());
      break;
    default:
      assert(0);  // increment and decrement shouldn't have a literal child.
    }
  }  
  return 0;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CUnaryExpression::Unparse(srcstream &sOut,
			  CCompileContext& context,
			  bool isStatement)
{
  CCompileError *error = 0;
  if (fSide == kPrefix) {
    sOut.ToLine(fLineNumber,false);
    switch(fOperator)
      {
      case INCR: sOut.out() << "++"; break;
      case DECR: sOut.out() << "--"; break;
      default  : sOut.out() << ((char) fOperator); break;
      }
  }
  if (fPrecedence <= fChild->GetPrecedence())
    sOut.ToLine(fLineNumber,false).out() << '(';
  error = fChild->Unparse(sOut,context);
  if (fPrecedence <= fChild->GetPrecedence())
    sOut.out() << ')';
  if (fSide == kPostfix) {
    switch(fOperator)
      {
      case INCR: sOut.out() << "++"; break;
      case DECR: sOut.out() << "--"; break;
      default  : sOut.out() << ((char) fOperator); break;
      }
  }
      
  return error;
}
