#include "find_leaks.h"
#include <algorithm>
#include <map>
#include "Expression.h"
#include "parser_decls.h"
#include "parser.h"  // for token values
#include "JavaFieldSignature.h"
#include "JavaClassFile.h"
#include "Compiler.h"
#include "CompileError.h"
#include "JavaMethodInfo.h"
#include "JavaCodeAttribute.h"
#include "CompileContext.h"
#include "Srcstream.h"
#include "Instantiator.h"
#include "BinaryExpression.h"
#include <iomanip.h>
#include <stdio.h>

static const unicode_string kLengthName = ::StringToUnicode("length");

//======================== CLocalVariableExpression ===========================

//
//  Method name : CLocalVariableExpression
//  Description : This constructor is used to make a variable name expression
//    using a field signature (a combination of type and name) and the index
//    of this variable as used in the local variable array.
//
CLocalVariableExpression::CLocalVariableExpression(
		       const CJavaFieldSignature& name, unsigned short index,
		       bool hasParameterName)
  : CExpression(name.GetType(), true),
    fName(name.GetFieldName()),
    fLocalVariableIndex(index),
    fPName(hasParameterName)
{
}

//
//  Method name : CLocalVariableExpression
//  Description : Copy constructor.
//
CLocalVariableExpression::CLocalVariableExpression(
				   const CLocalVariableExpression& source)
  : CExpression(source),
    fName(source.fName),
    fLocalVariableIndex(source.fLocalVariableIndex),
    fPName(source.fPName)
{
}

//
//  Method name : ~CLocalVariableExpression
//  Description : Destructor.
//
CLocalVariableExpression::~CLocalVariableExpression()
{
}

//
//  Method name : GetVariableIndex
//  Description : Returns the local variable index that will be used to hold
//    this variable in the virtual stack machine.
//
unsigned short
CLocalVariableExpression::GetVariableIndex() const
{
  return fLocalVariableIndex;
}

//
//  Method name : MayBeType
//  Description : Return type iff this expression can be interpreted 
//     (syntactically) as a type.
//
bool
CLocalVariableExpression::MayBeType() const
{
  return true;
}

//
//  Method name : ToType
//  Description : Returns a new CJavaTypeSignature for the type encoded
//     by this expression.  Returns NULL if this is impossible.
//
CJavaTypeSignature* 
CLocalVariableExpression::ToType() const
{
  return new CJavaTypeSignature(fName,0,0,fPName);
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CLocalVariableExpression::Unparse(srcstream &sOut,
				  CCompileContext& context,
				  bool isStatement)
{
  CCompileError* error = 0;
  if (!context.IsVariableInitialized(fLocalVariableIndex)) {
    unicode_string errorString = ::StringToUnicode("Local variable ");
    errorString += fName;
    errorString += ::StringToUnicode(" may not be initialized at this point");
    error = MakeError(errorString);
  } else {
    sOut.ToLine(fLineNumber,false).out() << ::SymbolToString(fName);
  }

  return error;
}

//
//  Method name : UnparseStore
//  Description : Writes the Java code for setting "this" to rValue to sOut.
//
CCompileError*
CLocalVariableExpression::UnparseStore(srcstream &sOut,
				       CCompileContext& context,
				       CExpression& rValue,
				       bool isStatement)
{
  CCompileError *error = 0;
  sOut.ToLine(fLineNumber,false).out() << ::SymbolToString(fName);  
  sOut.out() << " = ";
  error = rValue.Unparse(sOut,context);
  context.InitializeVariable(fLocalVariableIndex);
  return error;
}

//========================== CClassFieldExpression ============================

//
//  Method name : CClassFieldExpression
//  Description : Constructs an unknown class field out of its name only.
//
CClassFieldExpression::CClassFieldExpression(const unicode_string& label,
					     bool hasParameterName,
					     bool isDefault)
  : CExpression(CJavaTypeSignature::Void, true),
    fLabel(label),
    fClassAlias(0),
    fBase(0),
    fBaseType(0),
    fNeedCast(false),
    fNeedWrap(false),
    fHasPName(hasParameterName),
    fIsDefault(isDefault)
{
}

//
//  Method name : CClassFieldExpression
//  Description : Constructs an unknown class field out of a base type and
//     a field name. (e.g. ptype[Integer].staticField).
//
CClassFieldExpression::CClassFieldExpression(CJavaTypeSignature* adoptType,
					     const unicode_string& label)
  : CExpression(CJavaTypeSignature::Void, true),
    fLabel(label),
    fClassAlias(0),
    fBase(0),
    fBaseType(adoptType),
    fNeedCast(false),
    fNeedWrap(false),
    fHasPName(false),
    fIsDefault(false)
{  
}

//
//  Method name : CClassFieldExpression
//  Description : Constructs an unknown class field out of a base expresion
//      and a field name.
//
CClassFieldExpression::CClassFieldExpression(CExpression* adoptBase,
					     const unicode_string& label)
  : CExpression(CJavaTypeSignature::Void, true),
    fLabel(label),
    fClassAlias(0),
    fBase(adoptBase),
    fBaseType(0),
    fNeedCast(false),
    fNeedWrap(false),
    fHasPName(false),
    fIsDefault(false)
{
  if (fBase != 0)
    fLineNumber = fBase->GetLineNumber();
}

//
//  Method name : CClassFieldExpression
//  Description : Copy constructor.
//
CClassFieldExpression::CClassFieldExpression(
                                  const CClassFieldExpression& source)
  : CExpression(source),
    fLabel(source.fLabel),
    fFieldInfo(source.fFieldInfo),
    fClassAlias(source.fClassAlias),
    fBase(0),
    fBaseType(0),
    fNeedCast(source.fNeedCast),
    fNeedWrap(source.fNeedWrap),
    fHasPName(source.fHasPName),
    fIsDefault(source.fIsDefault)
{
  if (source.fBase != 0) {
    fBase = source.fBase->Clone();
  }
  if (source.fBaseType != 0) {
    fBaseType = new CJavaTypeSignature(*source.fBaseType);
  }
}

//
//  Method name : ~CClassFieldExpression
//  Description : Destructor.
//
CClassFieldExpression::~CClassFieldExpression()
{
  delete fBase;
  delete fBaseType;
}

//
//  Method name : IsStatic
//  Description : Returns true if this is a static field.
//
bool
CClassFieldExpression::IsStatic() const
{
  return fClassAlias != 0 && fFieldInfo.GetModifiers().fStatic != 0;
}

//
//  Method name : Resolve
//  Description : Attempts to resolve any ambiguities in this expression.
//      Does not attempt to resolve ambiguities in subexpressions.  On
//      failure, returns an error message.  If the expresion-class changes,
//      a new expression is written into setExpression.
//
CCompileError* 
CClassFieldExpression::Resolve(CCompileContext& context,
			       CExpression*& setExpression)
{
  if (fBase) {
    CAmbiguousAccess* base = DYNAMIC_CAST(CAmbiguousAccess, fBase);
    if (base) {
      CJavaTypeSignature type;
      CExpression* exp = 0;
      CCompileError* err = base->ResolveAE(context,exp,type);
      if (err) return err;
      if (exp) {
	delete fBase;
	fBase = exp;
      } else {
	delete fBase;
	fBase = 0;
	fBaseType = new CJavaTypeSignature(type);
      }
    }
  }
  return 0;
}

//
//  Method name : IsArrayLength
//  Description : Returns true if this field is used to get the length of an
//    array.
//
bool
CClassFieldExpression::IsArrayLength() const
{
  bool result = fBase != 0;
  if (result) {
    CJavaTypeSignature childType = fBase->GetType();
    result = childType.IsReference() && childType.IsArray() &&
      fLabel == kLengthName;
  }
  return result;
}

//
//  Method name : MayBeType
//  Description : Return type iff this expression can be interpreted 
//     (syntactically) as a type.
//
bool
CClassFieldExpression::MayBeType() const
{
  if (fBaseType || fIsDefault)
    return false;
  if (!fBase)
    return true;

  CLocalVariableExpression* lv = 
    DYNAMIC_CAST(CLocalVariableExpression, fBase);
  CClassFieldExpression* cf = 
    DYNAMIC_CAST(CClassFieldExpression, fBase);

  return (lv || (cf && fBase->MayBeType()));
}

//
//  Method name : ToType
//  Description : Returns a new CJavaTypeSignature for the type encoded
//     by this expression.  Returns NULL if this is impossible.
//
CJavaTypeSignature* 
CClassFieldExpression::ToType() const
{
  if (!MayBeType())
    return 0;
  if (!fBase)
    return new CJavaTypeSignature(fLabel, 0, 0, fHasPName);
   
  CJavaTypeSignature *baseType = fBase->ToType();
  unicode_string name;
  bool ret = baseType->GetBaseClassName(name);
  assert(ret);
  name += (unicode_char) '.';
  name += fLabel;

  delete baseType;
  return new CJavaTypeSignature(name);
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CClassFieldExpression::HandleEvaluateType(CCompileContext& context,
					  CJavaTypeSignature& setType)
{
  assert(fClassAlias == 0);
  CCompileError* error = 0;
  unicode_string::size_type lastSlash = fLabel.find_last_of('/');
  CInstantiator* ins = 0;
  if (fIsDefault) {
    fBaseType = new CJavaTypeSignature(fLabel,0,0,true);
    setType = *fBaseType;
    return 0;
  }
  if (fBase != 0) {
    ////
    // (expression).field
    ////
    assert(lastSlash == unicode_string::npos);
    error = EvaluateBaseType(context, setType);
  } else if (fBaseType != 0) {
    ////
    // Ptype[p1,p2].staticField
    ////
    assert(lastSlash == unicode_string::npos);
    CJavaTypeSignature classType;
    error = context.GetCompiler().CheckType(*fBaseType, context.GetClass(),
					    context.GetSource());
    if (error) error->SetLine(fLineNumber);
    if (!error) {
      *fBaseType = context.GetCompiler().FixType(*fBaseType,
						 context.GetSource());
      fClassAlias =
	context.GetCompiler().FindField(fLabel, *fBaseType,
					fFieldInfo, classType, 
					context.GetSource());
      if (fClassAlias == 0 || 
	  !context.GetCompiler().SameType(classType,*fBaseType,
					  context.GetSource())) {
	string errorString("No field ");
	errorString += ::UnicodeToUTF(fLabel);
	errorString += " found on class ";
	errorString += ::UnicodeToUTF(fBaseType->Disassemble());
	error = MakeError(errorString);
      } else {
	// Okay, class is good and we found a field.  Is it static?
	ins = new CInstantiator(*fClassAlias, classType.GetParameters());
	ins->Instantiate(setType,fFieldInfo.GetSignature().GetType());
	if (!IsStatic()) {
	  if (context.GetCompiler().DescendsFrom(context.GetClass().GetType(),
					      setType,context.GetSource())) {
	    fBase = new CSpecialExpression(CSpecialExpression::kThis);
	    fBase->InheritLineNumber(*this);
	  } else {
	    string errorString("Attempted to access non-static field ");
	    errorString += ::UnicodeToUTF(fLabel);
	    errorString += " from within static code";
	    error = MakeError(errorString);
	  }
	} else {
	  // It _is_ static!
	  fLValAlias = true;
	}
	fNeedWrap = setType.IsPrimitive() && 
	  fFieldInfo.GetSignature().GetType().IsParameter();
	fNeedCast = fFieldInfo.GetSignature().GetType().IsParameter() &&
	  !setType.IsParameter();
	delete ins;
      }
    }
  } else if (lastSlash == unicode_string::npos) {
    ////
    // field
    ////
    CJavaTypeSignature classType;
    fClassAlias = 
      context.GetCompiler().FindField(fLabel, 
				      context.GetClass().GetType(), 
				      fFieldInfo, classType,
				      context.GetSource());
    if (fClassAlias == 0) {
      string errorString("Unknown variable: ");
      errorString += ::UnicodeToUTF(fLabel);
      error = MakeError(errorString);
    } else {
      ins = new CInstantiator(*fClassAlias, classType.GetParameters());
      ins->Instantiate(setType,fFieldInfo.GetSignature().GetType());
      fNeedCast = fFieldInfo.GetSignature().GetType().IsParameter() &&
	!setType.IsParameter();
      fNeedWrap = fNeedCast && setType.IsPrimitive();
      if (IsStatic()) {	
	fBaseType = new CJavaTypeSignature(classType);
      } else {
	fBase = new CSpecialExpression(CSpecialExpression::kThis);
	fBase->InheritLineNumber(*this);
	if (context.InStaticCode()) {
	  string errorString("Attempted to access non-static field ");
	  errorString += ::UnicodeToUTF(fLabel);
	  errorString += " from within static code";
	  error = MakeError(errorString);
	}
      }
      delete ins;
    }
  } else {
    ////
    // name.name.field
    ////
    unicode_string baseName(fLabel, 0, lastSlash);
#ifdef OLDGCC
    fLabel.remove(0, lastSlash + 1);
#else
    fLabel.erase(0, lastSlash + 1);
#endif
    fBase = new CClassFieldExpression(baseName);
    fBase->InheritLineNumber(*this);
    error = EvaluateBaseType(context, setType);
    if (error != 0) {
      const CJavaClassFile* explicitClass =
	context.GetCompiler().LookupClass(baseName,context.GetSource());
      if (explicitClass != 0) {
	// className.field
	delete error;
	error = 0;
	delete fBase;
	fBase = 0;
	CJavaTypeSignature classType;
	fClassAlias =
	  context.GetCompiler().FindField(fLabel, 
					  explicitClass->GetType(), 
					  fFieldInfo, classType,
					  context.GetSource());
	if (fClassAlias == 0) {
	  string errorString("No field ");
	  errorString += ::UnicodeToUTF(fLabel);
	  errorString += " found on class ";
	  errorString += ::UnicodeToUTF(baseName);
	  error = MakeError(errorString);
	} else {
	  ins = new CInstantiator(*fClassAlias, classType.GetParameters());
	  ins->Instantiate(setType,fFieldInfo.GetSignature().GetType());
	  fNeedCast = fFieldInfo.GetSignature().GetType().IsParameter() &&
	    !setType.IsParameter();
	  fNeedWrap = fNeedCast && setType.IsPrimitive();
	  if (!IsStatic()) {
	    if (context.GetCompiler().DescendsFrom(
					       context.GetClass().GetType(),
					       setType,context.GetSource())) {
	      fBase = new CSpecialExpression(CSpecialExpression::kThis);
	      fBase->InheritLineNumber(*this);
	    } else {
	      string errorString("Attempted to access non-static field ");
	      errorString += ::UnicodeToUTF(fLabel);
	      errorString += " from within static code";	      
	      error = MakeError(errorString);
	    }
	  }
	  delete ins;
	}
      }
    }
  }
  ////
  // Okay: can we access the field?
  ////
  if (error == 0 && fClassAlias != 0) {
    if (fClassAlias != &context.GetClass() &&
	fFieldInfo.GetModifiers().fPublic == 0) {
      if (fFieldInfo.GetModifiers().fPrivate != 0) {
	string errorString("Invalid access to private field ");
	errorString += ::UnicodeToUTF(fLabel);
	error = MakeError(errorString);
      } else if (!(fClassAlias->GetPackageName() ==
		   context.GetClass().GetPackageName())) {
	// XXXXX This conditional is a bit of a hack.
	if (fFieldInfo.GetModifiers().fProtected == 0 ||
	    !context.GetCompiler().DescendsFrom(context.GetClass().GetType(),
						fClassAlias->GetType(),
						context.GetSource(),
						true)) { 
	  string errorString("Field ");
	  errorString += ::UnicodeToString(fLabel);
	  errorString += " is not accessible to this class.";
	  error = MakeError(errorString); 
	}
      }
    }
    if (error == 0) {
      SetType(setType);
      if (fFieldInfo.GetModifiers().fFinal != 0 &&
	  !context.GetCompiler().InClassInitializers()) {
	fLValue = false;
      }
    }
  }
  if (IsStatic() && fClassAlias->IsParameterized()) {
    fNeedWrap = setType.IsPrimitive() && 
      fFieldInfo.GetSignature().GetType().IsParameter();
    fNeedCast = !fFieldInfo.GetSignature().GetType().IsPrimitive();
    fLValAlias = true;
  }
  return error;
}

//
//  Method name : EvaluateBaseType
//  Description : This internal method is used by the field expression to
//    figure out the type of the fBase base expression that is being held,
//    and then use this to assess the type of this value, which is assigned
//    to the 'setType' parameter.
//  * If any problems arise during this evaluation, an error is created and
//    returned to the caller.
//
CCompileError*
CClassFieldExpression::EvaluateBaseType(CCompileContext& context,
					CJavaTypeSignature& setType)
{
  CCompileError* error = 0;
  assert(fBase != 0);
  CJavaTypeSignature childType;
  error = EvaluateType(fBase, context, childType, IsLValue());
  if (error == 0) {
    if (!childType.IsReference()) {
      string errorString("Cannot access field ");
      errorString += ::UnicodeToUTF(fLabel);
      errorString += " through non-reference value";
      error = MakeError(errorString);
    } else if (childType.IsArray()) {
      ////
      // x.length
      ///
      if (fLabel == kLengthName) {
	setType = CJavaTypeSignature::kInteger;
	fLValue = false;
      } else {
	error = MakeError("Invalid field access on array value.");
      }
    } else {
      unicode_string className;
      bool isClass = childType.GetBaseClassName(className);
      if (childType.IsParameter()) {
	string errorString("Request for field on parameter type ");
	errorString += ::UnicodeToUTF(className);
	error = MakeError(errorString);
      } else {
	assert(isClass);
	const CJavaClassFile* classFile =
	  context.GetCompiler().ClassOfType(childType,context.GetSource());
	if (classFile == 0) {
	  string errorString("Unknown class name: ");
	  errorString += ::UnicodeToUTF(className);
	  error = MakeError(errorString);
	} else {
	  CJavaTypeSignature classType;
	  fClassAlias =
	    context.GetCompiler().FindField(fLabel, childType,
					    fFieldInfo, classType,
					    context.GetSource());
	  if (fClassAlias == 0) {
	    string errorString("No field ");
	    errorString += ::UnicodeToUTF(fLabel);
	    errorString += " found on class ";
	    errorString += ::UnicodeToUTF(className);
	    error = MakeError(errorString);
	  } else {
	    CInstantiator ins(*fClassAlias, classType.GetParameters());
	    ins.Instantiate(setType,fFieldInfo.GetSignature().GetType());
	    fNeedCast = fFieldInfo.GetSignature().GetType().IsParameter() &&
	      !setType.IsParameter();
	    fNeedWrap = fNeedCast && setType.IsPrimitive();
	    if (!fBaseType)
	      fBaseType = new CJavaTypeSignature(classType);
	  }
	}
      }
    }
  }
  return error;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CClassFieldExpression::CreateConstantExpression(CExpression*& result,
						CCompileContext& context) const
{
  CCompileError* error = 0;
  result = 0;
  if (!IsArrayLength() && !fIsDefault) {
    error = context.GetCompiler().FindConstantField(result,
		 fClassAlias->GetClassName(), fFieldInfo.GetSignature(),
	         context, GetLineNumber());
  }
  return error;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CClassFieldExpression::Unparse(srcstream &sOut,
			       CCompileContext& context,
			       bool isStatement)
{
  CCompileError* error = 0;

  if (isStatement) fNeedCast = fNeedWrap = false; 
  
  if (fIsDefault) {
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::Default, *fBaseType);
    sOut.out() << "$W." << ::SymbolToString(auxMethod) << "()";
    return 0;
  }

  if (IsArrayLength() && 
      fBase->GetType().IsParameter() && 
      fBase->GetType().GetArrayBounds() == 1) {
    /////
    // T[] x; x.length;
    ////
    assert(context.GetClass().HasPInfo());
    CJavaTypeSignature param = fBase->GetType();
    param.SetArrayBounds(0);
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::ArrayLength, param);
    sOut.ToLine(fLineNumber,false);
    sOut.out() << "$W." << ::SymbolToString(auxMethod) << "(";
    error = fBase->Unparse(sOut,context);
    sOut.out() << ")";
    return error;
  }

  if (fNeedWrap) {
    sOut.ToLine(fLineNumber,false);
    sOut.out() << "((" << ::UnicodeToUTF(GetType().GetWrapperName()) << ")"; 
  } else if (fNeedCast) {
    sOut.ToLine(fLineNumber,false);
    sOut.out() << "((";
    GetType().Unparse(sOut);
    sOut.out() << ")";
  }

  if (IsArrayLength()) {
    ////
    // arry.length
    ////
    error = fBase->Unparse(sOut,context);
    sOut.out() << '.';
  } else if (IsStatic()) {
    ////
    // We have a static field
    ////
    sOut.ToLine(fLineNumber,false);
    assert(fClassAlias != 0 || fBaseType != 0);
    if (fBase != 0 && !fBaseType->IsParameterized()) {
      ////
      // NonParamExpression.staticField
      ////
      assert(fClassAlias != 0 && fBaseType != 0);
      error = fBase->Unparse(sOut,context);
      sOut.out() << '.';
    } else if (fBaseType && fBaseType->IsParameterized()) { 
      ////
      // PType.staticField
      ////
      const CJavaClassFile* baseClass = 
	context.GetCompiler().ClassOfType(*fBaseType,context.GetSource());
      assert(baseClass);
      CInstantiator ins(*baseClass, fBaseType->GetParameters());
      unicode_string auxMethod = baseClass->GetPInfo().GetAMM().
	AuxMethodName(CAuxMethodSignature::FetchStatic, fLabel);
      if (fFieldInfo.GetModifiers().fProtected && 
	  baseClass->GetPackageName() != context.GetClass().GetPackageName()) {
	// We need a $fetch_ method.
	sOut.out() << "$fetch_" << ::SymbolToString(fLabel) << "(";
	UnparseWClassCode(*fBaseType, context, sOut);
	sOut.out() << ")";
      } else {
	// We need an auxmethod.
	UnparseWClassCode(*fBaseType, context, sOut);
	sOut.out() << "." << ::UnicodeToUTF(auxMethod);
	if (fFieldInfo.GetModifiers().fPublic) {
	  sOut.out() << "()";
	} else {
	  sOut.out() << "(";
	  fBaseType->UnparseTokenClass(sOut);
	  sOut.out() << ".the_instance)";
	}
      }
      if (fNeedWrap) {
	sOut.out() << ").value";
      } else if (fNeedCast) {
	sOut.out() << ")";
      }
      return error;
    } else {
      ////
      // className.staticField
      ////
      CJavaTypeSignature rolledType;
      bool isThis = context.GetCompiler().RollType(rolledType, 
						   fClassAlias->GetType(), 
						   context);
      sOut.ToLine(fLineNumber,false);
      if (! isThis) {
	rolledType.Unparse(sOut);
	sOut.out() << ".";
      } 
    }
  } else {
    ////
    // (base.field)
    ////
    assert(fClassAlias != 0 && fBase != 0);
    error = fBase->Unparse(sOut,context);
    sOut.out() << '.';
  }
  sOut.ToLine(fLineNumber,false).out() << ::SymbolToString(fLabel);
  
  if (fNeedWrap) {
    sOut.out() << ").value";
  } else if (fNeedCast) {
    sOut.out() << ")";
  }

  return error;
}

//
//  Method name : UnparseStore
//  Description : Writes the Java code for setting "this" to rValue to sOut.
//
CCompileError* 
CClassFieldExpression::UnparseStore(srcstream &sOut,
				    CCompileContext& context,
				    CExpression& rValue,
				    bool isStatement)
{
  CCompileError* error = 0;
  if (fIsDefault) {
    return MakeError("Attempted to modify value of 'default'");
  }    
  if (IsStatic() && fBaseType != 0 && fBaseType->IsParameterized()) {
    ////
    // PType[T].staticField
    ////
    const CJavaClassFile* baseClass = 
      context.GetCompiler().ClassOfType(*fBaseType,context.GetSource());
    assert(baseClass);
    CInstantiator ins(*baseClass, fBaseType->GetParameters());
    fNeedWrap = (!fFieldInfo.GetSignature().GetType().IsPrimitive() &&
		 GetType().IsPrimitive());

    bool needCast = !isStatement && 
      fFieldInfo.GetSignature().GetType() != GetType();
    bool needUnwrap = !isStatement && 
      !fFieldInfo.GetSignature().GetType().IsPrimitive() &&
      GetType().IsPrimitive();

    if (needUnwrap) {
      sOut.out() << "((" << ::UnicodeToUTF(GetType().GetWrapperName()) << ")";
    } else if (needCast) {
      sOut.out() << "((";
      GetType().Unparse(sOut);
      sOut.out() << ")";
    }
    
    if (fFieldInfo.GetModifiers().fProtected &&
	baseClass->GetPackageName() != context.GetClass().GetPackageName()) {
      // We need a $set_ method.
      sOut.out() << "$set_" << ::SymbolToString(fLabel) << "( ";
      UnparseWClassCode(*fBaseType, context, sOut);
      sOut.out() << ", ";
      if (fNeedWrap)
	sOut.out() << "new " << ::UnicodeToUTF(GetType().GetWrapperName())
		   << "(";
      error = rValue.Unparse(sOut,context);
      if (fNeedWrap) 
	sOut.out() << ")";
      sOut.out() << ")";
    } else {
      // We need an auxmethod.
      unicode_string auxMethod = baseClass->GetPInfo().GetAMM().
	AuxMethodName(CAuxMethodSignature::SetStatic, fLabel);
      UnparseWClassCode(*fBaseType, context, sOut);
      sOut.out() << "." << ::UnicodeToUTF(auxMethod) << "( ";
      if (fNeedWrap)
	sOut.out() << "new " << ::UnicodeToUTF(GetType().GetWrapperName())
		   << "(";
      error = rValue.Unparse(sOut,context);
      if (fNeedWrap) 
	sOut.out() << ")";
      if (fFieldInfo.GetModifiers().fPublic) {
	sOut.out() << ")";
      } else {
	sOut.out() << ", ";
	fBaseType->UnparseTokenClass(sOut);
	sOut.out() << ".the_instance)";
      }
    }

    if (needUnwrap) {
      sOut.out() << ").value";
    } else if (needCast) {
      sOut.out() << ")";
    }

    return error;
  } else {
    // We don't need auxmethods .. just unparse normally.
    bool needWrap = fNeedWrap;
    fNeedWrap = fNeedCast = false;
    error = Unparse(sOut,context);
    if (error == 0) {
      sOut.out() << " = ";
      if (needWrap) {
	error = CExpression::UnparseWrap(GetType(), rValue, sOut, context);
      } else {
	error = rValue.Unparse(sOut, context);
      }
    }
  }
  return error;
}

//=========================== CArrayInitializer ===============================

//
//  Method name : CArrayInitializer
//  Description : Creates an array initializer expression from a list of
//    sub-expressions.
//
CArrayInitializer::CArrayInitializer(ExpressionList* adoptExpressions)
  : CExpression(CJavaTypeSignature::Void, false)
{
  if (adoptExpressions != 0) {
    fExpressions = *adoptExpressions;    
    if (!adoptExpressions->empty()) {
      fLineNumber = adoptExpressions->front()->GetLineNumber();
    }
    delete adoptExpressions;
  }
}

//
//  Method name : CArrayInitializer
//  Description : Copy constructor.
//
CArrayInitializer::CArrayInitializer(const CArrayInitializer& source)
  : CExpression(source)
{
  for (ExpressionList::const_iterator i = source.fExpressions.begin();
       !(i == source.fExpressions.end()); ++i) {
    fExpressions.push_back((*i)->Clone());
  }
}

//
//  Method name : ~CArrayInitializer
//  Description : Destructor
//
CArrayInitializer::~CArrayInitializer()
{
  for (ExpressionList::iterator i = fExpressions.begin();
       i != fExpressions.end(); i++) {
    delete *i;
  }
}

//
//  Method name : SetArrayType
//  Description : This method is used by the holder of an array initializer
//    to tell it what type it is.  This is backwards from all of the rest
//    of the expressions, which figure out their own type ... with
//    initializers, their type is imposed from the outside based on their
//    context.  This must be called before 'GenerateCode' can be used.
//
void
CArrayInitializer::SetArrayType(const CJavaTypeSignature& type)
{
  if (type.IsArray()) {
    SetType(type);
    CJavaTypeSignature childType = type;
    childType.SetArrayBounds(childType.GetArrayBounds() - 1);
    for (ExpressionList::iterator i = fExpressions.begin();
	 i != fExpressions.end(); ++i) {
      CArrayInitializer* child = DYNAMIC_CAST(CArrayInitializer, *i);
      if (child != 0) {
	child->SetArrayType(childType);
      }
    }
  }
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CArrayInitializer::Unparse(srcstream &sOut,
			   CCompileContext& context,
			   bool isStatement)
{
  CCompileError* error = 0;
  if (!GetType().IsArray()) {
    error = MakeError("Invalid non-array type for an array initializer.");
  } else {
    error = context.GetCompiler().CheckType(GetType(), context.GetClass(),
					    context.GetSource());
    if (error) error->SetLine(fLineNumber);
  }
  CJavaTypeSignature childType;
  CJavaTypeSignature elementType = GetType();
  elementType.SetArrayBounds(elementType.GetArrayBounds() - 1);
  if (!error && elementType.IsParameter() && 
      elementType.GetArrayBounds() == 0) {
    ExpressionList* expList = new ExpressionList;
    expList->push_front(new COrdinalLiteral((JavaSInt) fExpressions.size()));
    CExpression* dummyArray = new CNewArray(GetType(), expList);
    dummyArray->InheritLineNumber(*this);
    CJavaTypeSignature dummyType;
    error = EvaluateType(dummyArray, context, dummyType);
    if (!error)
      error = dummyArray->Unparse(sOut,context);
    delete dummyArray;
    return error;
  }
  sOut.ToLine(fLineNumber,false).out() << "{ ";
  for (ExpressionList::iterator i = fExpressions.begin();
       error == 0 && i != fExpressions.end(); ) {
    CArrayInitializer *ai = DYNAMIC_CAST(CArrayInitializer, *i);
    if (ai) {
      if (GetType().GetArrayBounds() == 1)
	error = MakeError("Type mismatch on array initializer");
    }
    error = EvaluateType(*i, context, childType);
    if (error == 0) {
      if (!context.GetCompiler().SameType(childType, elementType, 
					  context.GetSource())) {
	error = CCastExpression::CastPointer(*i, elementType, context);
      }
      if (error == 0) {
	error = (*i)->Unparse(sOut,context);
      }
    }
    ++i;
    if (i != fExpressions.end())
      sOut.out() << ", ";
  }
  
  sOut.out() << '}';
     
  return error;
}

//
//  Method name: UnparseAssignments
//  Description: Writes statements to initialize the elements of
//     "assignTo" to the expressions.  If allLevels, it calls array
//     creation expressions as well.
//  Requires: 'Unparse' has already been called, or allLevels is true.
//
CCompileError*
CArrayInitializer::UnparseAssignments(const CExpression *assignTo,
				      srcstream &sOut,
				      CCompileContext &context,
				      bool allLevels)
{
  CCompileError *error = 0;
  if (allLevels) {    
    ExpressionList* expList = new ExpressionList;
    expList->push_front(new COrdinalLiteral((JavaSInt) fExpressions.size()));
    CExpression* assignTo2 = assignTo->Clone();
    CNewArray* newArray = new CNewArray(GetType(), expList);
    CExpression* be = new CBinaryExpression(assignTo2, '=', newArray);
    CJavaTypeSignature t;
    error = EvaluateType(be, context, t);
    if (!error) {
      error = be->Unparse(sOut, context, true);
      sOut.out() << "; ";
    }
    delete be;
  }
  
  deque<CExpression*>::iterator exp = fExpressions.begin();
  for(unsigned int i = 0; !error && i < fExpressions.size(); ++i, ++exp) {
    COrdinalLiteral* index = new COrdinalLiteral((JavaSInt) i);
    CArrayIndex* access = new CArrayIndex(assignTo->Clone(), index);
    CJavaTypeSignature childType;
    CArrayInitializer *ai = DYNAMIC_CAST(CArrayInitializer, *exp);
    if (ai) {
      if (GetType().GetArrayBounds() == 1) 
	error = MakeError("Type mismatch on array initializer");
    }
    if (!error) {
      error = EvaluateType(*exp, context, childType);
    }
    if (!error) {
      if (ai) {
	error = ai->UnparseAssignments(access, sOut, context,allLevels);
	delete access;
      } else {
	CExpression* be = new CBinaryExpression(access, '=', (*exp)->Clone());
	CJavaTypeSignature t;
	error = EvaluateType(be, context, t);
	if (!error) {
	  be->Unparse(sOut,context,true);
	  sOut.out() << "; ";
	}
	delete be;
      }
    }
  }
  return error;
}

//=============================== CArrayIndex =================================

//
//  Method name : CArrayIndex
//  Description : Constructs an array index expression out of a base
//    expression and a subscript expression:
//         baseExpression [ subscriptExpression ]
//
CArrayIndex::CArrayIndex(CExpression* baseExpression, CExpression* index)
  : CExpression(CJavaTypeSignature::Void, true),
    fBase(baseExpression),
    fIndex(index)
{
  fLineNumber = fBase->GetLineNumber();
}

//
//  Method name : CArrayIndex
//  Description : Copy constructor.
//
CArrayIndex::CArrayIndex(const CArrayIndex& source)
  : CExpression(source), fBase(0), fIndex(0)
{
  if (source.fBase != 0) {
    fBase = source.fBase->Clone();
  }
  if (source.fIndex != 0) {
    fIndex = source.fIndex->Clone();
  }
}

//
//  Method name : ~CArrayIndex
//  Description : Destructor
//
CArrayIndex::~CArrayIndex()
{
  delete fBase;
  delete fIndex;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CArrayIndex::HandleEvaluateType(CCompileContext& context,
				CJavaTypeSignature& setType)
{
  assert(fBase != 0 && fIndex != 0);
  CJavaTypeSignature baseType, indexType;
  CCompileError* error = EvaluateType(fBase, context, baseType);
  if (error == 0) {
    error = EvaluateType(fIndex, context, indexType);
  }
  if (error == 0) {
    if (! baseType.IsArray()) {
      error = MakeError("Index taken of non-array typed expression.");
    } else if (!UnaryPromote(fIndex, context)) {
      error = MakeError("Non-integer expression used in array index.");
    } else {
      baseType.SetArrayBounds(baseType.GetArrayBounds() - 1);
      setType = baseType;
      if (baseType.IsParameter() && ! baseType.IsArray()) {
	fLValAlias = true;
      }
    }
  }
  return error;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CArrayIndex::Unparse(srcstream &sOut,
		     CCompileContext& context,
		     bool isStatement)
{
  CCompileError* error;
  if (GetType().IsParameter() && ! GetType().IsArray()) {
    ////
    // P[] x;  x[foo];
    ////
    // We need an auxmethod
    assert(context.GetClass().HasPInfo());
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::ArrayFetch,GetType());
    sOut.ToLine(fLineNumber).out()
      << "$W." << ::SymbolToString(auxMethod) << "(";
    error = fBase->Unparse(sOut,context);
    if (!error) {
      sOut.out() << ",";
      error = fIndex->Unparse(sOut,context);
      sOut.out() << ")";
    }
  } else { 
    // No auxmethods required.
    error = fBase->Unparse(sOut,context);
    sOut.out() << '[';
    if (error == 0)
      error = fIndex->Unparse(sOut,context);
    sOut.out() << ']';
  }
  return error;
}

//
//  Method name : UnparseStore
//  Description : Writes the Java code for setting "this" to rValue to sOut.
//
CCompileError* 
CArrayIndex::UnparseStore(srcstream &sOut,
			  CCompileContext& context,
			  CExpression& rValue,
			  bool isStatement)
{
  CCompileError* error = 0;
  if (GetType().IsParameter() && ! GetType().IsArray()) {
    ////
    // P[] x; x[foo] = bar;
    ////
    assert(context.GetClass().HasPInfo());
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::ArrayStore,GetType());   
    sOut.ToLine(fLineNumber).out()
      << "$W." << ::SymbolToString(auxMethod) << "(";
    error = fBase->Unparse(sOut,context);
    if (!error) {
      sOut.out() << ",";
      error = fIndex->Unparse(sOut,context);
      if (!error) {
	sOut.out() << ",";
	error = rValue.Unparse(sOut,context);
	sOut.out() << ")";
      }
    }
  } else {
    error = Unparse(sOut,context);
    sOut.out() << " = ";
    if (!error) {
      error = rValue.Unparse(sOut,context);
    }
  }
  return error;
}

//=============================== CMethodCall =================================

//
//  Method name : CMethodCall
//  Description : Constructs a method call expression node made up of
//    the target, method name and arguments to the method.
//
CMethodCall::CMethodCall(CExpression* adoptTarget,
			 const unicode_string& methodName,
			 ExpressionList* adoptArguments,
			 bool forceNonvirtual)
  : CExpression(CJavaTypeSignature::Void, false),
    fTarget(adoptTarget),
    fMethodName(methodName),
    fTargetType(0),
    fClassAlias(0),
    fMethodAlias(0),
    fConstraintMethodAlias(0),
    fForceNonvirtual(forceNonvirtual),
    fOnObject(false),
    fIsArrayClone(false),
    fIsPArrayClone(false)
{
  if (adoptArguments != 0) {
    fArguments = *adoptArguments;
    delete adoptArguments;
  }
  if (! fArguments.empty()) {
    fLineNumber = fArguments.front()->GetLineNumber();
  }
}

//
//  Method name : CMethodCall
//  Description : Constructs a method call expression node made up of
//    the method name and arguments to the method.
//
CMethodCall::CMethodCall(const unicode_string& methodName,
			 ExpressionList* adoptArguments,
			 bool forceNonvirtual)
  : CExpression(CJavaTypeSignature::Void, false),
    fTarget(0),
    fMethodName(methodName),
    fTargetType(0),
    fClassAlias(0),
    fMethodAlias(0),
    fConstraintMethodAlias(0),
    fForceNonvirtual(forceNonvirtual),
    fOnObject(false),
    fIsArrayClone(false),
    fIsPArrayClone(false)
{
  if (adoptArguments != 0) {
    fArguments = *adoptArguments;
    delete adoptArguments;
  }
  if (! fArguments.empty()) {
    fLineNumber = fArguments.front()->GetLineNumber();
  }
}

//
//  Method name : CMethodCall
//  Description : Constructs a method call expression node made up of
//    the target type, method name and arguments to the method.
//
CMethodCall::CMethodCall(CJavaTypeSignature* adoptTargetType,
			 const unicode_string& methodName,
			 ExpressionList* adoptArguments,
			 bool forceNonvirtual)
  : CExpression(CJavaTypeSignature::Void, false),
    fTarget(0),
    fMethodName(methodName),
    fTargetType(adoptTargetType),
    fClassAlias(0),
    fMethodAlias(0),
    fConstraintMethodAlias(0),
    fForceNonvirtual(forceNonvirtual),
    fOnObject(false),
    fIsArrayClone(false),
    fIsPArrayClone(false)
{
  if (adoptArguments != 0) {
    fArguments = *adoptArguments;
    delete adoptArguments;
  }
  if (! fArguments.empty()) {
    fLineNumber = fArguments.front()->GetLineNumber();
  }
}

//
//  Method name : CMethodCall
//  Description : Copy constructor.
//
CMethodCall::CMethodCall(const CMethodCall& source)
  : CExpression(source),
    fTarget(0),
    fMethodName(source.fMethodName),
    fTargetType(0),
    fClassAlias(source.fClassAlias),
    fMethodAlias(source.fMethodAlias),
    fConstraintMethodAlias(source.fConstraintMethodAlias),
    fForceNonvirtual(source.fForceNonvirtual),
    fOnObject(source.fOnObject),
    fIsArrayClone(source.fIsArrayClone),
    fIsPArrayClone(source.fIsPArrayClone)
{
  if (source.fTarget != 0) {
    fTarget = source.fTarget->Clone();
  }
  if (source.fTargetType != 0) {
    fTargetType = new CJavaTypeSignature(*source.fTargetType);
  }
  for (ExpressionList::const_iterator i = source.fArguments.begin();
       !(i == source.fArguments.end()); ++i) {
    fArguments.push_back((*i)->Clone());
  }
}

//
//  Method name : ~CMethodCall
//  Description : Destructor
//
CMethodCall::~CMethodCall()
{
  delete fTarget;
  delete fTargetType;
  for (ExpressionList::iterator i = fArguments.begin();
       i != fArguments.end(); i++) {
    delete *i;
  }
}

//  The name of the "equals" method.
static const unicode_string kEqualsName = ::StringToUnicode("equals");
static const unicode_string kCloneName = ::StringToUnicode("clone");

//
//  Method name : Resolve
//  Description : Attempts to resolve any ambiguities in this expression.
//      Does not attempt to resolve ambiguities in subexpressions.  On
//      failure, returns an error message.  If the expresion-class changes,
//      a new expression is written into setExpression.
//
CCompileError* 
CMethodCall::Resolve(CCompileContext& context,
		     CExpression*& setExpression)
{
  if (fTarget) {
    CAmbiguousAccess* base = DYNAMIC_CAST(CAmbiguousAccess, fTarget);
    if (base) {
      CJavaTypeSignature type;
      CExpression* exp = 0;
      CCompileError* err = base->ResolveAE(context,exp,type);
      if (err) return err;
      delete fTarget;
      if (exp) {
	fTarget = exp;
      } else {
	fTarget = 0;
	if (fTargetType) delete fTargetType;
	fTargetType = new CJavaTypeSignature(type);
      }
    }
  }
  return 0;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CMethodCall::HandleEvaluateType(CCompileContext& context,
				CJavaTypeSignature& setType)
{  
  CCompileError* error = 0;
  deque<CJavaTypeSignature> argumentTypes;
  CJavaTypeSignature resType; // which type is the method found on?
  // Find the types of the arguments
  for (ExpressionList::iterator i = fArguments.begin();
       error == 0 && i != fArguments.end(); i++) {
    CJavaTypeSignature argumentType;
    error = EvaluateType(*i, context, argumentType);
    argumentTypes.push_back(argumentType);
  }
  if (error != 0) return error;
  if (fTarget == 0 && fTargetType == 0) {
    ////
    // method(args)     or     name.method(args)
    ////
    unicode_string::size_type lastSlash = fMethodName.find_last_of('/');
    if (lastSlash == unicode_string::npos) {
      ////
      // method(args)
      ////
      string* matchErrorString = 0;
      pair<const CJavaClassFile*, const CJavaMethodInfo*>* match =
	context.GetCompiler().MatchMethod(fMethodName,
					  context.GetClass().GetType(), 
					  argumentTypes, 
					  matchErrorString, resType,
					  context.GetSource());
      if (match == 0) {
	if (matchErrorString == 0) {
	  string errorString("No matching method: ");
	  errorString += ::UnicodeToString(fMethodName);
	  errorString += '(';
	  deque<CJavaTypeSignature>::iterator arg = argumentTypes.begin();
	  while (arg != argumentTypes.end()) {
	    errorString += ::UnicodeToString((*arg).Disassemble());
	    if (++arg != argumentTypes.end()) {
	      errorString += ", ";
	    }
	  }
	  errorString += ") found on class ";
	  errorString +=
	    ::SymbolToString(context.GetClass().GetClassName());
	  if (fMethodName == context.GetClass().GetShortName())
	    errorString += " (Possibly missing a \"new\")";
	  error = MakeError(errorString);
	} else {
	  error = MakeError(*matchErrorString);
	  delete matchErrorString;
	}
      } else {
	// Successfully matched method(args);
	fClassAlias = match->first;
	if (fTargetType) delete fTargetType;
	fTargetType = new CJavaTypeSignature(resType);
	fMethodAlias = match->second;
	context.GetCompiler().WarnDeprecatedMethod(*fMethodAlias,
			          *fClassAlias, context, GetLineNumber());
	CInstantiator ins(*fClassAlias, resType.GetParameters());
	ins.Instantiate(setType,fMethodAlias->GetSignature().GetType());
	if (fMethodAlias->GetModifiers().fStatic != 0) {
	  fTarget = 0;
	  if (fTargetType) delete fTargetType;
	  fTargetType = new CJavaTypeSignature(resType);
	} else {
	  // Attempt to access non-static method...
	  if (context.InStaticCode()) {
	    string errorString = "Attempted to access non-static method "
	      + ::UnicodeToUTF(fMethodName) + " from within static code";
	    error = MakeError(errorString);				
	  } else {
	    fTarget = new CSpecialExpression(CSpecialExpression::kThis);
	    fTarget->InheritLineNumber(*this);
	    if (fTargetType) delete fTargetType;
	    fTargetType = new CJavaTypeSignature(resType);
	  }
	}
      }
      delete match;
    } else {
      ////
      // prefix.methodName(args)
      ////
      unicode_string prefix(fMethodName, 0, lastSlash);
      unicode_string shortName(fMethodName, lastSlash + 1);
      fMethodName = shortName;
      if (context.GetClass().HasParameterNamed(prefix)) {
	////
	// Parameter.methodName(args)
	////
	const CPJavaConstraint *cons = 
	  context.GetClass().GetPInfo().LookupConstraint(prefix);
	assert(cons);
	CJavaMethodSignature m(CJavaTypeSignature::kVoid,
			       fMethodName, argumentTypes);
	const CPJavaConstraint::ConstraintMethod* match =
	  context.GetCompiler().MatchConstraintMethod(m,*cons);
	if (match == 0) {
	  // No match for P.m(args)
	  unicode_string errorString = ::StringToUnicode("Parameter ");
	  errorString += prefix;
	  errorString += ::StringToUnicode(" has no matching method ");
	  errorString += shortName;
	  errorString += (unicode_char)'(';
	  for (deque<CJavaTypeSignature>::iterator i =
		 argumentTypes.begin(); !(i == argumentTypes.end());) {
	    errorString += (*i).Disassemble();
	    if (!(++i == argumentTypes.end())) {
	      errorString += (unicode_char)',';
	      errorString += (unicode_char)' ';
	    }
	  }
	  errorString += (unicode_char)')';
	  if (prefix == shortName)
	    errorString += 
	      ::StringToUnicode(" (Possibly missing a \"new\")");
	  error = MakeError(errorString);
	} else if (! match->fStatic) {
	  // Found P.virtual_m(args)
	  string errorString = "Invalid use of non-static parameter method ";
	  errorString += ::UnicodeToString(fMethodName);
	  error = MakeError(errorString);
	} else {
	  // Found P.static_m(args)
	  assert(match->fMethod);
	  setType = match->fMethod->GetType();
	  if (fTargetType) delete fTargetType;
	  fTargetType = new CJavaTypeSignature(prefix, 0, 0, true);
	  fTarget = 0;
	  fConstraintMethodAlias = match;
	}
      } else {
	////
	// ClassName.method(args)  or  Member.method(args)
	////
        error = EvaluatePrefixType(prefix, context);
        if (error == 0) {
          fTarget = new CClassFieldExpression(prefix);
          fTarget->InheritLineNumber(*this);
        } else {
          fTarget=0;
	  const CJavaClassFile* javaClass =
	    context.GetCompiler().LookupClass(prefix,context.GetSource());
	  if (javaClass != 0) {
	    // Check for bad instantiation.
	    CJavaTypeSignature t(prefix,0,0,false);
            if ( error ) delete error;
	    error = context.GetCompiler().CheckType(t, 
						    context.GetClass(),
						    context.GetSource());	  
	  }
	  if (javaClass != 0 && error == 0) {
            // ClassName.method(args)
	    string* matchErrorString = 0;
	    pair<const CJavaClassFile*, const CJavaMethodInfo*>* match =
	      context.GetCompiler().MatchMethod(fMethodName, 
					        javaClass->GetType(),
					        argumentTypes, 
					        matchErrorString, resType,
					        context.GetSource());
	    if (match == 0) {
	      if (matchErrorString == 0) {		
	        unicode_string errorString = ::StringToUnicode("Class ");
	        errorString += prefix;
	        errorString += ::StringToUnicode(" has no matching method ");
	        errorString += shortName;
	        errorString += (unicode_char)'(';
	        for (deque<CJavaTypeSignature>::iterator i =
		       argumentTypes.begin(); !(i == argumentTypes.end());) {
		  errorString += (*i).Disassemble();
		  if (!(++i == argumentTypes.end())) {
		    errorString += (unicode_char)',';
		    errorString += (unicode_char)' ';
		  }
	        }
	        errorString += (unicode_char)')';
	        error = MakeError(errorString);
	      } else {
	        error = MakeError(*matchErrorString);
	        delete matchErrorString;
	      }
	    } else {
	      // Found className.method(args)
	      fClassAlias = match->first;
	      fMethodAlias = match->second;
	      if (fTargetType) delete fTargetType;
	      fTargetType = new CJavaTypeSignature(resType);
	      context.GetCompiler().WarnDeprecatedMethod(*fMethodAlias,
						         *fClassAlias, 
						         context, 
						         GetLineNumber());
	      assert(fClassAlias != 0 && fMethodAlias != 0);
	      CInstantiator ins(*fClassAlias, resType.GetParameters());
	      ins.Instantiate(setType,fMethodAlias->GetSignature().GetType());
	      if (fMethodAlias->GetModifiers().fStatic == 0) {
	        string errorString = "Attempted to use non-static method ";
	        errorString += ::UnicodeToString(fMethodName);
	        errorString += " from within static code";
	        error = MakeError(errorString);
	      }
	    }
	  } else {
	    // we didn't find any class matching the prefix: try
	    //               class.field.method(args)
	    fTarget = new CClassFieldExpression(prefix);
	    fTarget->InheritLineNumber(*this);
          }
	}
      }
    }
  } else if (fTarget == 0 && fTargetType != 0) { 
    ////
    // PType[p1,p2].method(args)
    ////
    string* matchErrorString = 0;
    error = context.GetCompiler().CheckType(*fTargetType, 
					    context.GetClass(),
					    context.GetSource());    
    if (error) {
      error->SetLine(fLineNumber);
      return error;
    }
    *fTargetType = context.GetCompiler().FixType(*fTargetType,
						 context.GetSource());
    pair<const CJavaClassFile*, const CJavaMethodInfo*>* match = 0;
    match = context.GetCompiler().MatchMethod(fMethodName, 
					      *fTargetType,
					      argumentTypes, 
					      matchErrorString, resType,
					      context.GetSource());
    if (match == 0 && error == 0) {
      // Couldn't match PType[p1,p2].method(args)
      if (matchErrorString == 0) {
	unicode_string errorString = ::StringToUnicode("Class ");
	errorString += fTargetType->Disassemble();
	errorString += ::StringToUnicode(" has no matching method ");
	errorString += fMethodName;
	  errorString += (unicode_char)'(';
	  for (deque<CJavaTypeSignature>::iterator i =
		 argumentTypes.begin(); !(i == argumentTypes.end());) {
	    errorString += (*i).Disassemble();
	    if (!(++i == argumentTypes.end())) {
	      errorString += (unicode_char)',';
	      errorString += (unicode_char)' ';
	    }
	  }
	  errorString += (unicode_char)')';
	  error = MakeError(errorString);
      } else {
	error = MakeError(*matchErrorString);
	delete matchErrorString;
      }
    } else if (error == 0) {
      // matched PType[p1,p2].method(args)
      fClassAlias = match->first;
      fMethodAlias = match->second;
      if (fTargetType) delete fTargetType;
      fTargetType = new CJavaTypeSignature(resType);
      context.GetCompiler().WarnDeprecatedMethod(*fMethodAlias,
						 *fClassAlias, 
						 context, 
						 GetLineNumber());
      assert(fClassAlias != 0 && fMethodAlias != 0);
      if (fMethodAlias->GetModifiers().fStatic == 0) {
	string errorString = "Attempted to use non-static method ";
	errorString += ::UnicodeToString(fMethodName);
	errorString += " from within static code";
	error = MakeError(errorString);
      }
      CInstantiator ins(*fClassAlias, resType.GetParameters());
      ins.Instantiate(setType,fMethodAlias->GetSignature().GetType());
    }
  }
  CJavaTypeSignature targetType;
  if (error == 0 && fTarget != 0) {
    ////
    // (Expression).method()
    ////
    error = EvaluateType(fTarget, context, targetType);
  } 
  if (error == 0 && fTarget != 0 && fMethodAlias == 0) {
    // More (Expression).method()
    if (!targetType.IsReference()) {
      unicode_string errorString = ::StringToUnicode("Method ");
      errorString += fMethodName;
      errorString += ::StringToUnicode(" called in invalid type: ");
      errorString += targetType.Disassemble();
      error = MakeError(errorString);
    } else {
      // What _is_ the type of this expression?
      const CJavaClassFile* javaClass = 0;
      bool targetIsParameter = false;
      CJavaTypeSignature javaType = targetType;
      if (targetType.IsArray()) {
	// foo[] x; x.method();
	javaClass =
	  context.GetCompiler().LookupClass(CCompiler::kObjectName,0);
	javaType =
	  CJavaTypeSignature(CCompiler::kObjectName);
	if (fMethodName == kCloneName) {
	  fIsArrayClone = true;
	  if (targetType.IsParameter() && targetType.GetArrayBounds() == 1)
	    fIsPArrayClone = true;
	}
      } else {
	// Okay, it's not an array.  Is it a parameter?
	unicode_string className;
	bool isClass = targetType.GetBaseClassName(className);
	assert(isClass);
	if (context.GetClass().HasParameterNamed(className)) {
	  targetIsParameter = true;
	} else {
	  javaClass = context.GetCompiler().LookupClass(className,
							context.GetSource());
	}
      }
      if (javaClass == 0 && ! targetIsParameter) {
	// Couldn't find any class or parameter.
	unicode_string errorString = ::StringToUnicode("Method ");
	errorString += fMethodName;
	errorString += ::StringToUnicode(" called in invalid class: ");
	errorString += targetType.Disassemble();
	error = MakeError(errorString);
      } else if (targetIsParameter) {
	///
	// (parameter-valued-expression).method(args)
	////
	unicode_string className;
	bool isClass = targetType.GetBaseClassName(className);
	assert(isClass);
	assert(context.GetClass().HasPInfo());
	const CPJavaConstraint *cons = 
	  context.GetClass().GetPInfo().LookupConstraint(className);
	assert(cons);
	CJavaMethodSignature m(CJavaTypeSignature::kVoid,
			       fMethodName, argumentTypes);
	const CPJavaConstraint::ConstraintMethod* match =
	  context.GetCompiler().MatchConstraintMethod(m,*cons);
	bool universal = IsUniversal(context, argumentTypes, setType);
	if (match == 0 && !universal) {
	  // Didn't find method on constraint.  Is it universal?
	  string errorString("No matching method: ");
	  errorString += ::UnicodeToString(fMethodName);
	  errorString += '(';
	  deque<CJavaTypeSignature>::iterator arg = 
	    argumentTypes.begin();
	  while (arg != argumentTypes.end()) {
	    errorString += ::UnicodeToString((*arg).Disassemble());
	    if (++arg != argumentTypes.end()) {
	      errorString += ", ";
	    }
	  }
	  errorString += ") found on parameter ";
	  errorString += ::UnicodeToString(targetType.Disassemble());
	  error = MakeError(errorString);
	} else if (match == 0 && universal) {
	  assert(fOnObject);
	} else {
	  // We have a constraint method
	  if (!fOnObject) {
	    assert(match->fMethod);
	    setType = match->fMethod->GetType();
	    if (fTargetType) delete fTargetType;
	    fTargetType = new CJavaTypeSignature(className, 0, 0, true);
	    fConstraintMethodAlias = match;
	  }
	}
      } else {
	////
	// (non-parameter-valued-expression).method(args)
	////
	string* matchErrorString = 0;
	pair<const CJavaClassFile*, const CJavaMethodInfo*>* match =
	  context.GetCompiler().MatchMethod(fMethodName, 
					    javaType, argumentTypes, 
					    matchErrorString, resType,
					    context.GetSource());
	if (match == 0) {
	  if (matchErrorString == 0) {
	    string errorString("No matching method: ");
	    errorString += ::UnicodeToString(fMethodName);
	    errorString += '(';
	    deque<CJavaTypeSignature>::iterator arg = argumentTypes.begin();
	    while (arg != argumentTypes.end()) {
	      errorString += ::UnicodeToString((*arg).Disassemble());
	      if (++arg != argumentTypes.end()) {
		errorString += ", ";
	      }
	    }
	    errorString += ") found on class ";
	    errorString += ::UnicodeToString(targetType.Disassemble());
	    error = MakeError(errorString);
	  } else {
	    error = MakeError(*matchErrorString);
	    delete matchErrorString;
	  }
	} else {
	  fClassAlias = match->first;
	  fMethodAlias = match->second;
	  if (fTargetType) delete fTargetType;
	  fTargetType = new CJavaTypeSignature(resType);
	  context.GetCompiler().WarnDeprecatedMethod(*fMethodAlias,
				    *fClassAlias, context, GetLineNumber());
	  assert(fClassAlias != 0 && fMethodAlias != 0);
	  CInstantiator ins(*fClassAlias, resType.GetParameters());
	  ins.Instantiate(setType,fMethodAlias->GetSignature().GetType());
	  delete match;
	}
      }
    }
  }
  // XXXX fix:  stackUsed is being set based on the original argument sizes,
  // not the sizes after casting to match the formal parameters, so passing
  // integer 5 to a parameter taking a double is only going to count for 1
  // spot, not 2.
  if (error == 0 && (fMethodAlias != 0 || fConstraintMethodAlias != 0))  {
    CJavaMethodSignature sig;
    if (fMethodAlias) {
      assert(fClassAlias);
      CInstantiator ins(*fClassAlias, resType.GetParameters());
      ins.Instantiate(sig,fMethodAlias->GetSignature());
    } else {
      sig = *fConstraintMethodAlias->fMethod;
    }
    assert(sig.ParameterCount() == fArguments.size());
    deque<CJavaTypeSignature>::const_iterator parameter =
      sig.ParametersBegin();
    for (ExpressionList::iterator argument = fArguments.begin();
	 error == 0 && argument != fArguments.end(); ++argument, ++parameter) {
      if (!context.GetCompiler().SameType((*argument)->GetType(),
					  *parameter,context.GetSource())) {
	error = CCastExpression::CastPointer(*argument, *parameter, context);
      }
    }
  }
  return error;
}

//
//  Method name : EvaluatePrefixType
//  Description : This internal method is used by the method expression to
//    figure out the type of the prefix expression that is being held
//  * If any problems arise during this evaluation, an error is created and
//    returned to the caller.
//
CCompileError*
CMethodCall::EvaluatePrefixType(unicode_string prefix,
                                CCompileContext& context)
{
  CCompileContext newContext(context);

  CJavaTypeSignature targetType;
  
  CExpression* newFTarget = new CClassFieldExpression(prefix);
  newFTarget->InheritLineNumber(*this);
  
  CCompileError* error = EvaluateType( newFTarget, newContext, targetType );
  delete newFTarget;

  return error;
}

//
//  Method name : IsUniversal
//  Description : Returns true iff this method is one of those which may
//      be called on all parameter types without being declared in the where
//      clause.
//    * Sets fOnObject to true iff this method is provided by java.lang.Object,
//      and re-sets fClassAlias, fMethodAlias, and fTargetType as appropriate.
//
bool
CMethodCall::IsUniversal(CCompileContext &context,
			 const deque<CJavaTypeSignature> &argumentTypes,
			 CJavaTypeSignature &setType)
{
  static const unicode_string kHashCodeName = ::UTFToUnicode("hashCode");
  static const unicode_string kToStringName = ::UTFToUnicode("toString");

  // Is this method on java.lang.Object?
  string* matchErrorString = 0;
  CJavaTypeSignature resType;
  pair<const CJavaClassFile*, const CJavaMethodInfo*>* JLOmatch =
    context.GetCompiler().MatchMethod(fMethodName, 
			   CJavaTypeSignature(CCompiler::kObjectName), 
				      argumentTypes,
				      matchErrorString, 
				      resType,
				      context.GetSource());
  delete matchErrorString;
  // Don't try this optimization with "equals()".
  fOnObject = (JLOmatch != 0 && 
	       JLOmatch->second->GetSignature().GetName() != kEqualsName &&
	       JLOmatch->second->GetSignature().GetName() != kCloneName);  
  if (fOnObject) {    
    fClassAlias = JLOmatch->first;
    fMethodAlias = JLOmatch->second;    
    if (fTargetType) delete fTargetType;
    fTargetType = new CJavaTypeSignature(resType);
    context.GetCompiler().WarnDeprecatedMethod(*fMethodAlias,
		            *fClassAlias, context, GetLineNumber());
    assert(fClassAlias != 0 && fMethodAlias != 0);
    CInstantiator ins(*fClassAlias, resType.GetParameters());
    ins.Instantiate(setType,fMethodAlias->GetSignature().GetType());
    delete JLOmatch;
  }

  // Now, return the right value.
  return (argumentTypes.size() == 0 &&
	  ((fMethodName == kHashCodeName) ||
	   (fMethodName == kToStringName)));
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CMethodCall::Unparse(srcstream &sOut,
		     CCompileContext& context,
		     bool isStatement)
{
  assert((fMethodAlias != 0 && fClassAlias != 0) || 
	 fConstraintMethodAlias != 0);
  assert(fTargetType != 0);   
  CCompileError* error = context.GetCompiler().CheckType(*fTargetType,
							 context.GetClass(),
							 context.GetSource());
  if (error) {
    error->SetLine(fLineNumber);
    return error;
  }
  bool constructor = (fMethodName == CCompiler::kConstructorName);

  // Check visibilitiy
  if (fClassAlias != &context.GetClass() &&
      fMethodAlias && fMethodAlias->GetModifiers().fPublic == 0) {
    if (fMethodAlias->GetModifiers().fPrivate != 0) {
      error = MakeError("Invalid access to private method");
    } else if (!(fClassAlias->GetPackageName() ==
	    context.GetClass().GetPackageName())) {
      if (fMethodAlias->GetModifiers().fProtected != 0 &&
	  !context.GetCompiler().DescendsFrom(context.GetClass().GetType(),
					      fClassAlias->GetType(),
					      context.GetSource(),
					      true)) {
	// array.clone is not protected.
	if (! fIsArrayClone) {
	  string errorString("Method ");
	  errorString += ::UnicodeToString(fMethodName);
	  errorString += " is not accessible to this class.";
	  error = MakeError(errorString);
	}
      }
    }
  } 
  
  // Check throws
  if (fMethodAlias && fMethodAlias->ThrowsExceptions() && !fIsArrayClone) {
    for (deque<unicode_string>::const_iterator i = fMethodAlias->ThrowsBegin();
       error == 0 && !(i == fMethodAlias->ThrowsEnd()); ++i) {
      if (!context.Throwable(CJavaTypeSignature(*i))) {
	string message = ::SymbolToString(*i);
	message += " must be caught or declared in "
	  "this method's 'throws' list.";
	error = MakeError(message);
	return error;
      }
    }
  } else if (fConstraintMethodAlias && 
	     fConstraintMethodAlias->fThrows->size() > 0 && !fIsArrayClone) {
    for (deque<CJavaTypeSignature>::iterator i = 
	   fConstraintMethodAlias->fThrows->begin(); 	  
	 error == 0 && i != fConstraintMethodAlias->fThrows->end(); ++i) {
      if (!context.Throwable(*i)) {
	unicode_string message = (*i).Disassemble();
	message += ::UTFToUnicode(" must be caught or declared in this "
				  "method's 'throws' list.");
	error = MakeError(message);	
	return error;
      }
    }
  }

  bool castReturn = false;
  bool unwrapReturn = false;
  deque<CJavaTypeSignature>::const_iterator argument;
  deque<CJavaTypeSignature>::const_iterator transArgument;
  CJavaMethodSignature insMethod; // Used for argument
  CJavaMethodSignature transMethod; // Used for transArgument
  int methodidx = 0;

  // Special case: clone array of Param
  if (fIsPArrayClone) {
    // Get Where {$clone_array}
    static const unicode_string kCloneArrayName = 
      ::UTFToUnicode("$clone_array");
    CJavaMethodSignature ms = fMethodAlias->GetSignature();
    assert(fTarget);
    CJavaTypeSignature targetType = fTarget->GetType();
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::Where,
		    targetType,
		    & kCloneArrayName,
		    &ms,
		    0,
		    CAuxMethodSignature::Virtual);
    sOut.out() << "$W." << ::SymbolToString(auxMethod) << "(";
    fTarget->Unparse(sOut, context);
    sOut.out() << ")";
    return error;
  }

  if (! fConstraintMethodAlias) {
    // We have a real method.
    assert(fMethodAlias);
    CInstantiator ins(*fClassAlias, fTargetType->GetParameters());
    ins.Instantiate(insMethod,fMethodAlias->GetSignature());
    argument = insMethod.ParametersBegin();
    transArgument = fMethodAlias->GetTranslatedSignature().ParametersBegin();
    CJavaTypeSignature transReturn = 
      fMethodAlias->GetTranslatedSignature().GetType();
    if (transReturn != GetType() && !isStatement) {
      // The method returns a parameter type, so we need to cast or unwrap 
      // the return value.
      castReturn = true;
      sOut.out() << "((";
      if (GetType().IsPrimitive() && !transReturn.IsPrimitive()) {
	unwrapReturn = true;
	sOut.out() << ::UnicodeToUTF(GetType().GetWrapperName());
      } else {
	GetType().Unparse(sOut);
      }
      sOut.out() << ")";
    }
    if (fMethodAlias->GetModifiers().fStatic != 0) {
      // Static method
      assert(fClassAlias->GetAccessFlags().fInterface == 0);
      assert(fTargetType);
      sOut.ToLine(fLineNumber);
      CJavaTypeSignature rolledType;
      bool isThis =
	context.GetCompiler().RollType(rolledType, *fTargetType, context);
      if (! isThis) {
	rolledType.Unparse(sOut);
	sOut.out() << '.';
      }
    } else {
      // Virtual method -- may be an explicit constructor call.
      assert(fTarget != 0);
      CSpecialExpression *maybeThis = DYNAMIC_CAST(CSpecialExpression, 
						   fTarget);
      if (maybeThis == 0 || constructor ||
	  maybeThis->GetType() != CSpecialExpression::kThis) {
	bool needParens = fTarget->GetPrecedence() > 0 && !constructor;
	if (needParens) sOut.out() << "(";
	error = fTarget->Unparse(sOut,context);
	if (needParens) sOut.out() << ")";
	if (!constructor)
	  sOut.out() << '.';
      }
    }
    // We have a real method, and we've unparsed its base.
    // Now unparse the name.
    if (fClassAlias->HasPInfo()) {
      // Do we need to mangle the name?
      assert(fClassAlias->GetPInfo().GetMethods());
      MethodNameMap::const_iterator found = fClassAlias->GetPInfo().
	GetMethods()->find(fMethodAlias->GetSignature());
      if (found != fClassAlias->GetPInfo().GetMethods()->end())
	methodidx = (*found).second.fIndex;
    }

    if (! constructor) {
      sOut.ToLine(fLineNumber,false).out() 
	<< ::SymbolToString(fMethodName);
      if (methodidx > 0) 
	sOut.out() << "$" << methodidx;
    }    
    sOut.out() << '(';
    // Now for the arguments.
    if ((fMethodAlias->GetModifiers().fStatic || constructor) && 
	fClassAlias->IsParameterized()) {
      // We need to add a where object as the first argument.
      ++transArgument;
      UnparseWClassCode(*fTargetType, context, sOut);
      // Should only be done if there is an arg!
      if (! fArguments.empty()) {
        sOut.out() << ",";
      }
    }
  } else {
    ////
    // We're calling a constraint method.
    ////
    argument = fConstraintMethodAlias->fMethod->ParametersBegin();
    transMethod = *fConstraintMethodAlias->fMethod;
    transArgument = transMethod.ParametersBegin();

    // Do we need to unwrap the return?
    if (fConstraintMethodAlias->fMethod->GetType().IsParameter() &&
	!GetType().IsParameter() && !isStatement) {
      castReturn = true; 
      sOut.out() << "((";
      GetType().Unparse(sOut);
      sOut.out() << ")";
      if (GetType().IsPrimitive())
	unwrapReturn = true;
    } 
    assert(fTargetType);
    unicode_string constraintMethodName = 
      fConstraintMethodAlias->fMethod->GetName();
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::Where,
		    *fTargetType,
		    & constraintMethodName,
		    fConstraintMethodAlias->fMethod,
		    fConstraintMethodAlias->fThrows, 
		    (fConstraintMethodAlias->fStatic ?
		     CAuxMethodSignature::Static : 
		     CAuxMethodSignature::Virtual));
    sOut.out() << "$W." << ::SymbolToString(auxMethod) << "(";
    if (!fConstraintMethodAlias->fStatic) {
      assert(fTarget);
      error = fTarget->Unparse(sOut,context);
      if (fConstraintMethodAlias->fMethod->ParameterCount() > 0)
	sOut.out() << ", ";
    }
  }

  // Now write out the arguments
  for (ExpressionList::iterator i = fArguments.begin();
       error == 0 && i != fArguments.end(); ) {
    // error = (*i)->Unparse(sOut, context);	  
    error = CExpression::UnparseArgument(*argument,
					 *transArgument,
					 *(*i), sOut, context);
    ++i;
    ++argument;
    ++transArgument;
    if (i != fArguments.end())
      sOut.out() << ", ";	
  } 
  if (constructor) {
    if (methodidx>0 && fArguments.size()>0)
      sOut.out() << ", ";      
    for(int i = 0; i < methodidx; ++i) {
      sOut.out() << (i < methodidx-1) ? "0, " : "0";
    }
  }

  sOut.out() << ")";
  if (castReturn)
    sOut.out() << ")";
  if (unwrapReturn)
    sOut.out() << ".value";
  return error;
}

//=========================== CSpecialExpression ==============================

//
//  Method name : CSpecialExpression
//  Description : Constructs a dopey hard-coded special expression value.
//
CSpecialExpression::CSpecialExpression(CSpecialExpression::Type type)
  : CExpression(CJavaTypeSignature::Void, false),
    fType(type)
{
}

//
//  Method name : CSpecialExpression
//  Description : Copy constructor.
//
CSpecialExpression::CSpecialExpression(const CSpecialExpression& source)
  : CExpression(source),
    fType(source.fType)
{
}

//
//  Method name : ~CSpecialExpression
//  Description : Destructor
//
CSpecialExpression::~CSpecialExpression()
{
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CSpecialExpression::HandleEvaluateType(CCompileContext& context,
				       CJavaTypeSignature& setType)
{
  CCompileError* error = 0;
  switch (fType) {
  case kThis:
    setType = context.GetClass().GetType();
    break;
  case kSuper:
    if (context.GetClass().GetSuperclassName().size() == 0) {
      error = MakeError("Invalid use of 'super' in Object.");
    }
    if (context.GetClass().HasPInfo()) {
      assert(context.GetClass().GetPInfo().GetSuper());
      setType = *context.GetClass().GetPInfo().GetSuper();
    } else { 
      setType = CJavaTypeSignature(context.GetClass().GetSuperclassName());
    }
    break;
  default:
    assert(0);
    break;
  }
  return error;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CSpecialExpression::Unparse(srcstream &sOut,
			    CCompileContext& context,
			    bool isStatement)
{
  sOut.ToLine(fLineNumber,false);
  switch (fType)
    {
    case kThis  : sOut.out() << "this"; break;
    case kSuper : sOut.out() << "super"; break;
    default     : assert(0); break;
    }

  return 0;
}

//============================== CInstanceof ==================================

//
//  Method name : CInstanceof
//  Description : Constructs an expression to model the instanceof operator
//
CInstanceof::CInstanceof(CExpression* adoptBase,
			 const CJavaTypeSignature& toType)
  : CExpression(CJavaTypeSignature::Boolean, false),
    fBase(adoptBase),
    fToType(toType)
{
  fPrecedence = 5;
}

//
//  Method name : CInstanceof
//  Description : Copy constructor.
//
CInstanceof::CInstanceof(const CInstanceof& source)
  : CExpression(source), fBase(0), fToType(source.fToType)
{
  if (source.fBase != 0) {
    fBase = source.fBase->Clone();
  }
}

//
//  Method name : ~CInstanceof
//  Description : Destructor
//
CInstanceof::~CInstanceof()
{
  delete fBase;
}


//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CInstanceof::Unparse(srcstream &sOut,
		     CCompileContext& context,
		     bool isStatement)
{
  CJavaTypeSignature baseType;
  CCompileError* error = EvaluateType(fBase, context, baseType);
  
  if (error == 0 && !fToType.IsReference()) {
    error = MakeError("instanceof requires a reference type.");
  }
  if (error == 0) {
    fToType = context.GetCompiler().FixType(fToType,context.GetSource());
    error = context.GetCompiler().CheckType(fToType,context.GetClass(),
					    context.GetSource());    
    if (error) error->SetLine(fLineNumber);
  } 
  if (error == 0) {
    if (baseType.IsArray()) {
      error = MakeError("Invalid instanceof check against array type.");
    } else if (!fToType.IsReference()) {
      unicode_string errorString = fToType.Disassemble();
      errorString += ::StringToUnicode(" is not a reference type.");
      error = MakeError(errorString);
    } else if (!baseType.IsReference()) {
      unicode_string errorString = baseType.Disassemble();
      errorString += ::StringToUnicode(" is not a reference type.");
      error = MakeError(errorString);
    } else if (!context.GetCompiler().CastableType(baseType, fToType,
						   context.GetSource())) {
      unicode_string errorString = fToType.Disassemble();
      errorString += ::StringToUnicode(" is not an assignable subtype of ");
      errorString += baseType.Disassemble();
      error = MakeError(errorString);
    }
  }
  if (error != 0) return error;
  context.GetCompiler().WarnUnenforcedCast(fToType,context,
					   GetLineNumber());
  if (!fToType.IsParameter() && (!fToType.IsParameterized() ||
				  fToType.IsArray())) {
    ////
    // No auxmethods needed.
    ////
    error = fBase->Unparse(sOut,context);
    sOut.out() << " instanceof ";
    fToType.Unparse(sOut);
  } else if (fToType.IsParameter()) { 
    // Use an Instanceof auxmethod. 
    unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
      AuxMethodName(CAuxMethodSignature::InstanceOf, fToType);
    sOut.out()
      << (isStatement ? "$W." : "($W.")
      << ::UnicodeToUTF(auxMethod)
      << "((Object) ";
    error = fBase->Unparse(sOut,context);
    sOut.out() << (isStatement ?  ")" : "))" );
  } else {
    // Call an $instanceof method.
    assert(fToType.IsParameterized() && ! fToType.IsParameter());
    fToType.UnparseWType(sOut);
    sOut.out() << ".$instanceof(";
    error = fBase->Unparse(sOut,context);
    for (deque<CJavaTypeSignature>::const_iterator i = 
	   fToType.GetParameters()->begin();
	 i != fToType.GetParameters()->end(); ++i) {
      if ((*i).IsInstantiable()) {
	sOut.out() << ", \"" << (*i).Compile() << "\"";
      } else {
	unicode_string auxMethod = context.GetClass().GetPInfo().GetAMM().
	  AuxMethodName(CAuxMethodSignature::PSig, (*i));
	sOut.out() << ", $W." << ::UnicodeToUTF(auxMethod) << "()";
      }
    }
    sOut.out() << ")";
    if (fToType.IsInstantiable()) {
      CJavaTypeSignature t = 
	context.GetCompiler().FixType(fToType,context.GetSource());
      context.GetCompiler().GetIList().AddType(t,context.GetSource());
    }
  }
  return error;
}

//========================= CTypeOrExpression ============================

//
//  Constructor
//  Description: Create a new TypeOrExpression from the provided
//     expression.
//
CTypeOrExpression::CTypeOrExpression(CExpression *adoptExp)
  : fIsType(false)
{
  fVal.e = adoptExp;
  assert(adoptExp);
}

//
//  Constructor
//  Description: Create a new TypeOrExpression from the provided
//     type.
//
CTypeOrExpression::CTypeOrExpression(CJavaTypeSignature *adoptType)
  : fIsType(true)
{
  fVal.t = adoptType;
  assert(adoptType);
}

//
//  Copy constructor
//
CTypeOrExpression::CTypeOrExpression(const CTypeOrExpression &source)
  : fIsType(source.fIsType)
{
  if (fIsType) {
    fVal.t = new CJavaTypeSignature(*source.fVal.t);
  } else {
    fVal.e = source.fVal.e->Clone();
  }
}

//
//  Method name: GetExpression 
//  Description: Return the expression encapsulated as this.
//
CExpression*
CTypeOrExpression::GetExpression() const
{
  assert(! fIsType);
  return fVal.e;
}

//
//  Method name: GetType
//  Description: Return the type encapsulated as this.
//
CJavaTypeSignature*
CTypeOrExpression::GetType() const
{
  assert(fIsType);
  return fVal.t;
}

//
//  Destructor
//
CTypeOrExpression::~CTypeOrExpression()
{
  if (fIsType)
    delete fVal.t;
  else
    delete fVal.e;
}


//========================= CAmbiguousAccess =============================

//
//  Constructor
//  Description: Create a new AmbiguousAccess expression from the
//     provided name and expression list.
//
CAmbiguousAccess::CAmbiguousAccess(unicode_string* adoptName, 
				   CExpression* adoptBase,
				   TypeOrExpList* adoptEList,
				   bool hasPName)
  : CExpression(CJavaTypeSignature::Void, true),
    fBaseString(adoptName),
    fBaseExpression(adoptBase),
    fEList(adoptEList),
    fHasPName(hasPName)
{
  assert(adoptName);
  assert(adoptEList);
  assert(adoptEList->size() > 0);
}

//
//  Copy constructor
//
CAmbiguousAccess::CAmbiguousAccess(const CAmbiguousAccess& source)
  : CExpression(source),
    fHasPName(source.fHasPName)
{
  fBaseString = new unicode_string(* source.fBaseString);
  fBaseExpression = source.fBaseExpression->Clone();
  fEList = new TypeOrExpList;
  for (TypeOrExpList::iterator i = source.fEList->begin();
       i != source.fEList->end(); ++i) {
    fEList->push_back( new CTypeOrExpression(*(*i)) );
  }
}

//
//  Destructor
//
CAmbiguousAccess::~CAmbiguousAccess()
{
  delete fBaseString;
  delete fBaseExpression;
  for (TypeOrExpList::iterator i = fEList->begin();
       i != fEList->end(); ++i) {
    delete (*i);
  }
  delete fEList;
}

//  
//  Method name : MayBeType
//  Description : Returns true iff this expression may be (from syntactic 
//     considerations only) a type.
//
bool
CAmbiguousAccess::MayBeType() const
{
  // Rationale: Syntactically, any expression of the form of 
  //    name[expOrType...] _may_ be a type iff every member of
  //    expOrType may be a type.
  assert(fEList);
  for (TypeOrExpList::iterator i = fEList->begin();
       i != fEList->end(); ++i) {
    if ((! (*i)->IsType()) && (! (*i)->GetExpression()->MayBeType()))
      return false;
  }
  return true;
}
//
//  Method name : MayBeIndex
//  Description : Returns true iff this expression may be (from syntactic
//     considerations only) an array index.
//
bool
CAmbiguousAccess::MayBeIndex() const
{
  // Rationale: Syntactically, any expression of the form of
  //   name[expOrType...] _may_ be an array index iff there is exactly
  //   one member of 'expOrType', and that member is _not_ itself a type.
  assert(fEList);
  if (fEList->size() != 1)
    return false;
  TypeOrExpList::const_iterator i = fEList->begin();
  if ((*i)->IsType())
    return false;
  return true;
}

//
//  Method name : ToType
//  Description : Returns a new CJavaTypeSignature for the type encoded
//     by this AmbiguousAccess.  Returns NULL if this is impossible.
//
CJavaTypeSignature *
CAmbiguousAccess::ToType() const
{
  deque<CJavaTypeSignature> *params = new deque<CJavaTypeSignature>;
  for (TypeOrExpList::iterator i = fEList->begin();
       i != fEList->end(); ++i) {
    if ((*i)->IsType()) {
      assert((*i)->GetType());
      params->push_back(* ((*i)->GetType()));
    } else {
      CJavaTypeSignature *t = (*i)->GetExpression()->ToType();
      if (! t) {
	delete params;
	return 0;
      }
      params->push_back(*t);
      delete t;
    }
  }
  CJavaTypeSignature *type =  new CJavaTypeSignature(*fBaseString,
						     0,
						     params);
  return type;
}

//
//  Method name : ToIndex
//  Description : Returns a new CArrayIndex encoded by this AmbiguousAccess.
//     Returns NULL if this is impossible
//
CArrayIndex *
CAmbiguousAccess::ToIndex() const
{
  assert(fEList);
  if (fEList->size() != 1)
    return 0;
  TypeOrExpList::const_iterator i = fEList->begin();
  if ((*i)->IsType())
    return 0;

  CArrayIndex* exp = new CArrayIndex(fBaseExpression->Clone(),
				     (*i)->GetExpression()->Clone());

  exp->InheritLineNumber(*this);
  return exp;
}

//
//  Method name : ToTypeAndDimension
//
//
CCompileError* 
CAmbiguousAccess::ToTypeAndDimension(CJavaTypeSignature& setType,
				     CExpression*& setExpression)
{
  assert(fEList);
  if (fEList->size() != 1)
    return MakeError("Found list where expression expected");
 
  TypeOrExpList::const_iterator i = fEList->begin();
  if ((*i)->IsType())
    return MakeError("Found type where expression expected");
 
  setType = CJavaTypeSignature(*fBaseString, 0, 0, fHasPName);
  setExpression = (*i)->GetExpression()->Clone();
  return 0;
}

//
//  Method name : ResolveAE
//  Description : Attempts to resolve the ambiguous expression as either a
//      type or an array index.  On failure, returns an error message,
//      failure, returns an error message.  On success, sets setExpression
//      to a newly allocated, resolved expression if the original expression
//      is an ArrayIndex, but sets setType to the proper type otherwise.
//    * This method should only be called when either interpretation is
//      "acceptable": in the cases of CNewArray, CMethodCall, and 
//       CClassFieldExpression.  Otherwise 'Resolve' will be called, forcing
//       an 'ArrayIndex' reading.
//
CCompileError* 
CAmbiguousAccess::ResolveAE(CCompileContext& context,
			    CExpression*& setExpression,
			    CJavaTypeSignature& setType)
{
  // Rationale: If fBaseExpression is a valid  field or local variable,
  //   then we should insist on a local variable access, always.  Otherwise,
  //   we should insist on a type.
  CJavaTypeSignature childType;
  assert(fBaseExpression);
  CCompileError* err = 0;
  if (MayBeIndex()) {
    err = EvaluateType(fBaseExpression,
		       context,
		       childType,
		       false);    
    if (err == 0) {
      CExpression* exp = ToIndex();
      if (!exp) {
	unicode_string str = ::StringToUnicode("Expected to find array index");
	return MakeError(str);
      }
      setExpression = exp;
      return 0;
    } 
  }

  CJavaTypeSignature* t = ToType();
  if (!t) {
    unicode_string str = ::StringToUnicode("Expected to find type");
    return MakeError(str);
  }
  CCompileError* err2 = context.GetCompiler().CheckType(*t,
							context.GetClass(),
							context.GetSource());
  if (err2) {
    err2->SetLine(fLineNumber);
    delete err;
    return err2;
  }
  setType = *t;
  delete t;
  delete err;
  return 0;
}

//
//  Method name : Resolve
//  Description : Attempts to resolve any ambiguities in this expression.
//      Does not attempt to resolve ambiguities in subexpressions.  On
//      failure, returns an error message.  On success, sets setExpression
//      to a newly allocated, resolved expression if the original expression
//      is ambiguous; leaves setExpression alone otherwise.
//
CCompileError* 
CAmbiguousAccess::Resolve(CCompileContext& context,
			  CExpression*& setExpression)
{
  CExpression* exp = ToIndex();
  if (!exp) {
    unicode_string str = ::StringToUnicode("Bad array-index expression");
    return MakeError(str);
  }
  setExpression = exp;
  return 0;
}

