// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: VariableDeclaration.C,v 1.20 1998/10/09 19:21:31 nickm Exp $
#pragma implementation
#include "find_leaks.h"

#include "VariableDeclaration.h"
#include "Expression.h"
#include "CompileError.h"
#include "JavaMethodInfo.h"
#include "JavaClassFile.h"
#include "Compiler.h"
#include "BinaryExpression.h"
#include "CompileContext.h"
#include "Srcstream.h"
#include "dynamic_cast.h"

//
//  Method name : CVariableDeclaration
//  Description : Constructs a variable declaration out of its name, type, and
//    optional initializer.
//
CVariableDeclaration::CVariableDeclaration(const CJavaTypeSignature& type,
 	         const unicode_string& name, CExpression* adoptInitializer)
  : fSignature(type, name),
    fInitializer(adoptInitializer),
    fLineNumber(yylineno)
{
  if (fInitializer != 0) {
    fLineNumber = fInitializer->GetLineNumber();
  }
}

//
//  Method name : ~CVariableDeclaration
//  Description : Destructor
//
CVariableDeclaration::~CVariableDeclaration()
{
  delete fInitializer;
}

//
//  Method name : GetType
//  Description : Returns the type of this variable.
//
CJavaTypeSignature
CVariableDeclaration::GetType() const
{
  return fSignature.GetType();
}

//
//  Method name : SetType
//  Description : Sets the type of this variable.
//
void
CVariableDeclaration::SetType(const CJavaTypeSignature& type)
{
  fSignature = CJavaFieldSignature(type, fSignature.GetFieldName());
}

//
//  Method name : Unparse
//  Description : Write the Java code for this variable declaration to sOut.
//
CCompileError*
CVariableDeclaration::Unparse(srcstream &sOut,
			      CCompileContext &context)
{
  CCompileError* error = 0;
  CJavaTypeSignature type = fSignature.GetType();   
  error = context.GetCompiler().CheckType(type, 
					  context.GetClass(),
					  context.GetSource());
  if (error) { error->SetLine(fLineNumber); }
  sOut.ToLineF(fLineNumber,false);    
  if (error == 0) {
    CJavaFieldSignature checked(type, fSignature.GetFieldName());
    checked.Unparse(sOut);
  }
  if (fInitializer && !error) {
    error = CExpression::EvaluateType(fInitializer, context, type);
    CBinaryExpression* binaryExpression =
      DYNAMIC_CAST(CBinaryExpression,fInitializer);
    if (error == 0) {
      assert(binaryExpression != 0 && binaryExpression->fRight != 0);
      CArrayInitializer *arrayInitializer = 
	DYNAMIC_CAST(CArrayInitializer,binaryExpression->fRight);
      if (!error && arrayInitializer != 0 && type.IsParameter()) {
	sOut.out() << "; ";
	error = arrayInitializer->UnparseAssignments(binaryExpression->fLeft,
						     sOut,context,true);
      } else if (!error) {
	sOut.out() << " = ";
	error = binaryExpression->fRight->Unparse(sOut, context); 	
      }
    }
    CLocalVariableExpression *lve = 
      DYNAMIC_CAST(CLocalVariableExpression,binaryExpression->fLeft);
    if (lve)
      context.InitializeVariable(lve->GetVariableIndex());      
  }
  sOut.out() << "; ";
  return error;
}

//
//  Method name : GetConstantValue
//  Description : This method is used to try to extract a constant literal
//    value associated with this variable declaration's initializer.  If
//    no initializer exists, or if the initializer can't be resolved to
//    a constant value, then the intoPointer is set to 0, otherwise it is
//    filled with the appropriate constant value.
//    If a problem occurs during assessment, an error is created and returned
//    to the caller, otherwise 0 is returned.
//
CCompileError*
CVariableDeclaration::GetConstantValue(CExpression*& intoPointer,
				       CCompileContext& context)
{
  CCompileError* error = 0;
  intoPointer = 0;
  if (fInitializer != 0) {
    CJavaTypeSignature type;
    error =
      CExpression::EvaluateType(fInitializer, context, type, false);
    CBinaryExpression* initializer =
      DYNAMIC_CAST(CBinaryExpression, fInitializer);
    if (initializer != 0 && error == 0) {
      const CExpression* initializerValue = initializer->GetRightArgument();
      assert(initializerValue != 0);
      error = initializerValue->CreateConstantExpression(intoPointer, context);
    }
  }
  return error;
}

//
// Method name : UnparseInitializer
// Description : Generate the code necessary to initialize this variable,
//    _without_ the declaration.
//
CCompileError* 
CVariableDeclaration::UnparseInitializer(srcstream &sOut,
					 CCompileContext &context)
{
  CJavaTypeSignature type;  
  if (!fInitializer) return 0;
  CCompileError* error = 
    CExpression::EvaluateType(fInitializer, context, type, false);

  CBinaryExpression* binaryExpression =
    DYNAMIC_CAST(CBinaryExpression,fInitializer);
  assert(binaryExpression != 0 && binaryExpression->fRight != 0);
  CArrayInitializer *arrayInitializer = 
    DYNAMIC_CAST(CArrayInitializer,binaryExpression->fRight);  

  if (!error && !arrayInitializer) {
    error = fInitializer->Unparse(sOut,context,true);
    sOut.out() << "; ";
  }
  if (!error && arrayInitializer != 0 && type.IsParameter()) {
    error = arrayInitializer->UnparseAssignments(binaryExpression->fLeft,
						 sOut,context,true);
  }
  return error;
}

//
// Method name : UnparseNoInitializer
// Description : Generate the code necessary to declare this variable,
//    _without_ the initializer.
//
CCompileError* 
CVariableDeclaration::UnparseNoInitializer(srcstream &sOut,
					   CCompileContext &context)
{
  CJavaFieldSignature t = 
    context.GetCompiler().FixFieldSignature(fSignature, context.GetSource());
  CCompileError *error = context.GetCompiler().CheckType(t.GetType(), 
							 context.GetClass(),
							 context.GetSource());
  t.Unparse(sOut);  
  sOut.out() << "; ";
  return error;
}

